from __future__ import annotations

import pathlib

import procrunner
import pytest

from dxtbx.serialize import load

from dials.array_family import flex
from dials.command_line.ssx_integrate import run_integration, working_phil
from dials.util.options import ArgumentParser

# Note that tests are grouped and run serially, to stop many processes trying to
# extract data from images at same time, which appears to lead to race
# conditions in some CI jobs.


@pytest.mark.xdist_group(name="group1")
def test_ssx_integrate_fullprocess(dials_data, tmp_path):
    # Download data set and the internally referenced images
    ssx = dials_data("cunir_serial_processed", pathlib=True)
    dials_data("cunir_serial", pathlib=True)
    pathlib.Path.mkdir(tmp_path / "nuggets")
    result = procrunner.run(
        [
            "dials.ssx_integrate",
            ssx / "indexed.refl",
            ssx / "indexed.expt",
            "nproc=1",
            "batch_size=3",
            "output.json=data.json",
            "output.nuggets=nuggets",
            "algorithm=stills",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("integrated_1.refl").is_file()
    assert tmp_path.joinpath("integrated_1.expt").is_file()
    assert tmp_path.joinpath("integrated_2.refl").is_file()
    assert tmp_path.joinpath("integrated_2.expt").is_file()
    assert tmp_path.joinpath("dials.ssx_integrate.html").is_file()
    assert tmp_path.joinpath("data.json").is_file()
    for i in range(1, 6):
        assert tmp_path.joinpath(f"nuggets/nugget_integrated_{i}.json").is_file()


@pytest.mark.xdist_group(name="group1")
def test_ssx_integrate_fullprocess_ellipsoid(dials_data, tmp_path):
    # Download data set and the internally referenced images
    ssx = dials_data("cunir_serial_processed", pathlib=True)
    dials_data("cunir_serial", pathlib=True)

    expts = load.experiment_list(ssx / "indexed.expt", check_format=False)
    expts[3:4].as_file(tmp_path / "single.expt")
    refls = flex.reflection_table.from_file(ssx / "indexed.refl")
    refls = refls.select_on_experiment_identifiers([expts[3].identifier])
    refls.as_file(tmp_path / "single.refl")

    result = procrunner.run(
        [
            "dials.ssx_integrate",
            tmp_path / "single.refl",
            tmp_path / "single.expt",
            "nproc=1",
            "algorithm=ellipsoid",
            "n_macro_cycles=2",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("integrated_1.refl").is_file()
    assert tmp_path.joinpath("integrated_1.expt").is_file()
    assert tmp_path.joinpath("dials.ssx_integrate.html").is_file()
    expts = load.experiment_list(tmp_path / "integrated_1.expt", check_format=False)
    mosaicity = expts[0].profile.mosaicity()
    assert mosaicity["radial"] < 0.0250 and mosaicity["radial"] > 0.0240
    assert mosaicity["angular_0"] < 0.0265 and mosaicity["angular_0"] > 0.0255
    assert mosaicity["angular_1"] < 0.006 and mosaicity["angular_1"] > 0.004


@pytest.mark.parametrize("algorithm,expected_n_refls", [("stills", 614)])
@pytest.mark.xdist_group(name="group1")
def test_ssx_integrate_algorithms(dials_data, algorithm, expected_n_refls):
    # Download data set and the internally referenced images
    ssx = dials_data("cunir_serial_processed", pathlib=True)
    dials_data("cunir_serial", pathlib=True)

    indexed_refl = flex.reflection_table.from_file(
        ssx / "indexed.refl"
    ).split_by_experiment_id()
    indexed_expts = load.experiment_list(ssx / "indexed.expt", check_format=True)

    parser = ArgumentParser(phil=working_phil, check_format=False)
    params, _ = parser.parse_args(args=[], quick_parse=True)

    params.algorithm = algorithm
    params.nproc = 1
    params.image_range = "1:2"

    results = list(run_integration(indexed_refl, indexed_expts, params))
    assert len(results) == 1  # only one batch expected
    experiments = results[0][0]
    reflections = results[0][1]

    assert len(experiments) == 2
    assert len(reflections) == pytest.approx(expected_n_refls, abs=9)



# Example code for how to load experiments and reflections in the DIALS


from __future__ import annotations

from libtbx.phil import parse

import dials.util.options
from dials.util import show_mail_handle_errors

help_message = """

pass experiment.expt indexed.refl
"""

phil_scope = parse(
    """
    png = 'example.png'
      .type = str
      .help = 'Output name for .png'
    """,
    process_includes=True,
)


class Script:
    """A class for running the script."""

    def __init__(self):
        usage = "dials.experiment_data [options] indexed.expt indexed.refl"

        self.parser = dials.util.options.ArgumentParser(
            usage=usage,
            phil=phil_scope,
            epilog=help_message,
            check_format=False,
            read_experiments=True,
            read_reflections=True,
        )

    def run(self):
        from scitbx import matrix

        params, options = self.parser.parse_args(show_diff_phil=True)

        experiments = dials.util.options.flatten_experiments(params.input.experiments)
        reflections = dials.util.options.flatten_reflections(params.input.reflections)

        if len(experiments) == 0:
            self.parser.print_help()
            return

        if len(reflections) != 1:
            self.parser.print_help()
            return

        reflections = reflections[0]

        print(f"Read {len(reflections)} reflections")

        indexed = reflections.select(reflections.get_flags(reflections.flags.indexed))

        print(f"Kept {len(indexed)} indexed reflections")

        for name in sorted(indexed.keys()):
            print(f"Found column {name}")

        for reflection in indexed[:3]:
            print(reflection)

        # verify that these experiments correspond to exactly one imageset, one
        # detector, one beam (obviously)
        for experiment in experiments[1:]:
            assert experiment.imageset == experiments[0].imageset
            assert experiment.beam == experiments[0].beam
            assert experiment.detector == experiments[0].detector

        # now perform some calculations - the only things different from one
        # experiment to the next will be crystal models
        print("Crystals:")
        for experiment in experiments:
            print(experiment.crystal)
        detector = experiments[0].detector
        beam = experiments[0].beam
        imageset = experiments[0].imageset

        # derived quantities
        wavelength = beam.get_wavelength()
        s0 = matrix.col(beam.get_s0())

        print(f"Wavelength: {wavelength:g}Å")
        print(f"Beam vector s0:\n{s0}")

        # in here do some jiggery-pokery to cope with this being interpreted as
        # a rotation image in here i.e. if scan is not None; derive goniometer
        # matrix else set this to identity

        scan = experiments[0].scan
        goniometer = experiments[0].goniometer

        if scan and goniometer:
            angle = scan.get_angle_from_array_index(
                0.5 * sum(imageset.get_array_range())
            )
            axis = matrix.col(goniometer.get_rotation_axis_datum())
            F = matrix.sqr(goniometer.get_fixed_rotation())
            S = matrix.sqr(goniometer.get_setting_rotation())
            R = S * axis.axis_and_angle_as_r3_rotation_matrix(angle, deg=True) * F
        else:
            R = matrix.sqr((1, 0, 0, 0, 1, 0, 0, 0, 1))

        print(f"Rotation matrix:\n{R}")

        assert len(detector) == 1


if __name__ == "__main__":
    with show_mail_handle_errors():
        script = Script()
        script.run()



from __future__ import annotations

from unittest import mock

import pytest

from libtbx import phil

from dials.algorithms.scaling.cross_validation.cross_validate import cross_validate
from dials.algorithms.scaling.cross_validation.crossvalidator import (
    DialsScaleCrossValidator,
)
from dials.util.options import ArgumentParser


def generated_param():
    """Generate a param phil scope."""
    phil_scope = phil.parse(
        """
      include scope dials.command_line.scale.phil_scope
  """,
        process_includes=True,
    )
    parser = ArgumentParser(phil=phil_scope, check_format=False)
    parameters, _ = parser.parse_args(args=[], quick_parse=True, show_diff_phil=False)
    return parameters


def test_crossvalidator():
    """Use the dials.scale cross validator to test the general methods of the
    parent class."""
    experiments = []
    reflections = []

    # test the create_results_dict method
    crossvalidator = DialsScaleCrossValidator(experiments, reflections)
    crossvalidator.create_results_dict(1)
    assert len(crossvalidator.results_dict) == 1
    assert (
        len(crossvalidator.results_dict[0])
        == len(crossvalidator.results_metadata["names"]) + 1
    )
    crossvalidator = DialsScaleCrossValidator(experiments, reflections)
    crossvalidator.create_results_dict(2)
    assert len(crossvalidator.results_dict) == 2
    assert (
        len(crossvalidator.results_dict[0])
        == len(crossvalidator.results_metadata["names"]) + 1
    )
    assert (
        len(crossvalidator.results_dict[1])
        == len(crossvalidator.results_metadata["names"]) + 1
    )

    # now test the add configuration
    keys = ("model",)
    values = (["a", "b"],)
    crossvalidator.set_results_dict_configuration(keys, values)
    assert crossvalidator.results_dict[0]["configuration"][0] == "model=a"
    assert crossvalidator.results_dict[1]["configuration"][0] == "model=b"

    # now try adding results to dict
    results_1 = [1.0, 1.5, 0.5, 3.0, 4.0, 1.0]
    results_2 = [1.0, 2.0, 1.0, 3.0, 2.0, 2.0]
    results_3 = [1.0, 2.0, 1.0, 3.0, 3.0, 0.0]
    results_4 = [1.0, 2.0, 1.0, 3.0, 3.5, 0.5]
    crossvalidator.add_results_to_results_dict(0, results_1)
    assert crossvalidator.results_dict[0]["work Rmeas"] == [1.0]
    assert crossvalidator.results_dict[0]["free Rmeas"] == [1.5]
    assert crossvalidator.results_dict[0]["work CC1/2"] == [3.0]
    assert crossvalidator.results_dict[0]["free CC1/2"] == [4.0]
    # add some more so that can test interpreting results
    crossvalidator.add_results_to_results_dict(0, results_2)
    crossvalidator.add_results_to_results_dict(1, results_3)
    crossvalidator.add_results_to_results_dict(1, results_4)

    # Now try interpreting results - check that values are calculated correctly
    st = crossvalidator.interpret_results()
    r1 = ["model=a", "mean", "1.0", "1.75*", "0.75*", "3.0", "3.0", "1.5"]
    r2 = ["", "std dev", "0.0", "0.25", "0.25", "0.0", "1.0", "0.5"]
    r3 = ["model=b", "mean", "1.0", "2.0", "1.0", "3.0", "3.25*", "0.25*"]
    r4 = ["", "std dev", "0.0", "0.0", "0.0", "0.0", "0.25", "0.25"]
    assert st._rows[0] == r1
    assert st._rows[1] == r2
    assert st._rows[2] == r3
    assert st._rows[3] == r4


def test_dialsscalecrossvalidator():
    """Test the methods of the dials.scale cross validator"""
    experiments = []
    reflections = []

    def mock_script():
        script = mock.MagicMock()
        script.scaler.work_free_stats = [1.0, 2.0, 3.0, 4.0]
        return script

    # test get results from script
    crossvalidator = DialsScaleCrossValidator(experiments, reflections)
    script = mock_script()
    results = crossvalidator.get_results_from_script(script)
    assert results == [1.0, 2.0, 3.0, 4.0]

    params = generated_param()
    params.scaling_options.free_set_percentage = 20.0
    # test get free set offset
    fsp = crossvalidator.get_free_set_percentage(params)
    assert fsp == 20.0

    # test set free set offset
    params = crossvalidator.set_free_set_offset(params, 5)
    assert params.scaling_options.free_set_offset == 5

    # test get/set parameters
    assert crossvalidator.get_parameter_type("model") == "choice"
    assert crossvalidator.get_parameter_type("physical.absorption_correction") == "bool"
    assert crossvalidator.get_parameter_type("physical.lmax") == "int"
    assert crossvalidator.get_parameter_type("physical.decay_interval") == "float"

    params = crossvalidator.set_parameter(params, "model", "physical")
    assert params.model == "physical"
    params = crossvalidator.set_parameter(params, "physical.decay_interval", 50.0)
    assert params.physical.decay_interval == 50.0
    params = crossvalidator.set_parameter(params, "physical.lmax", 10)
    assert params.physical.lmax == 10
    params = crossvalidator.set_parameter(
        params, "weighting.error_model.error_model", None
    )
    assert params.weighting.error_model.error_model is None
    params = crossvalidator.set_parameter(params, "cut_data.d_min", 1.8)
    assert params.cut_data.d_min == 1.8
    params = crossvalidator.set_parameter(params, "scaling_options.outlier_zmax", 7.53)
    assert params.scaling_options.outlier_zmax == 7.53
    with pytest.raises(ValueError):
        _ = crossvalidator.set_parameter(params, "bad_parameter", 7.53)

    # defer testing of run_script to command line tests


def test_cross_validate_script():
    """Test the script, mocking the run_script and interpret results calls"""

    param = generated_param()
    crossvalidator = DialsScaleCrossValidator([], [])

    # test expected error raise due to unspecified parameter
    param.cross_validation.cross_validation_mode = "multi"
    with pytest.raises(ValueError):
        cross_validate(param, crossvalidator)

    # test single mode
    param.cross_validation.cross_validation_mode = "single"
    param.cross_validation.nfolds = 2
    fpath = "dials.algorithms.scaling.cross_validation."
    with mock.patch(
        fpath + "crossvalidator.DialsScaleCrossValidator.run_script"
    ) as mock_run_script:
        with mock.patch(
            fpath + "crossvalidator.DialsScaleCrossValidator.interpret_results"
        ) as mock_interpret:
            cross_validate(param, crossvalidator)
            assert mock_run_script.call_count == 2
            assert mock_interpret.call_count == 1

    # test multi mode
    param = generated_param()
    param.cross_validation.cross_validation_mode = "multi"
    param.cross_validation.nfolds = 2
    fpath = "dials.algorithms.scaling.cross_validation."
    with mock.patch(
        fpath + "crossvalidator.DialsScaleCrossValidator.run_script"
    ) as mock_run_script:
        with mock.patch(
            fpath + "crossvalidator.DialsScaleCrossValidator.interpret_results"
        ) as mock_interpret:
            param.cross_validation.parameter = "physical.absorption_correction"
            cross_validate(param, crossvalidator)
            assert mock_run_script.call_count == 4
            assert mock_interpret.call_count == 1

            param.cross_validation.parameter = "physical.decay_interval"
            with pytest.raises(ValueError):
                cross_validate(param, crossvalidator)

            param.cross_validation.parameter = "physical.absorption_correction"
            param.cross_validation.parameter_values = ["True", "False"]
            cross_validate(param, crossvalidator)
            assert mock_run_script.call_count == 8
            assert mock_interpret.call_count == 2

            param.cross_validation.parameter = "physical.decay_interval"
            param.cross_validation.parameter_values = ["5.0", "10.0"]
            cross_validate(param, crossvalidator)
            assert mock_run_script.call_count == 12
            assert mock_interpret.call_count == 3

            param.cross_validation.parameter = "model"
            param.cross_validation.parameter_values = ["array", "physical"]
            cross_validate(param, crossvalidator)
            assert mock_run_script.call_count == 16
            assert mock_interpret.call_count == 4

            param.cross_validation.parameter = "physical.lmax"
            param.cross_validation.parameter_values = ["4", "6"]
            cross_validate(param, crossvalidator)
            assert mock_run_script.call_count == 20
            assert mock_interpret.call_count == 5

            param.cross_validation.parameter = "bad_interval"
            with pytest.raises(ValueError):
                cross_validate(param, crossvalidator)

            param.cross_validation.cross_validation_mode = "bad"
            with pytest.raises(ValueError):
                cross_validate(param, crossvalidator)



from __future__ import annotations

import logging
from typing import Any, List

import pytest

import dials.util.log
from dials.util.mp import batch_multi_node_parallel_map, multi_node_parallel_map


def test_LoggingContext():

    # configure logging
    dials.util.log.config(verbosity=2)

    # get some loggers
    idx_logger = logging.getLogger("dials.algorithms.indexing")
    dials_logger = logging.getLogger("dials")

    # check the logging level is as expected
    assert idx_logger.getEffectiveLevel() == logging.DEBUG
    assert dials_logger.getEffectiveLevel() == logging.DEBUG

    # selectively change logging level and check this has worked
    with dials.util.log.LoggingContext(idx_logger, logging.ERROR):
        assert idx_logger.getEffectiveLevel() == logging.ERROR
        assert dials_logger.getEffectiveLevel() == logging.DEBUG

    # check logging levels are as they were before
    assert idx_logger.getEffectiveLevel() == logging.DEBUG
    assert dials_logger.getEffectiveLevel() == logging.DEBUG

    # now check we can pass logger as a string
    with dials.util.log.LoggingContext("dials.algorithms.indexing", logging.WARNING):
        assert idx_logger.getEffectiveLevel() == logging.WARNING
        assert dials_logger.getEffectiveLevel() == logging.DEBUG

    # check logging levels are as they were before
    assert idx_logger.getEffectiveLevel() == logging.DEBUG
    assert dials_logger.getEffectiveLevel() == logging.DEBUG


def log_something(_: Any) -> List[logging.LogRecord]:
    """
    Create a dummy info log record.

    A little dummy function to pass to the dials.util.mp parallel map functions,
    which simply logs a single info message.

    Args:
        _:  Absorb the expected argument and do nothing with it.

    Returns:
        The log records created during the calling of this function.
    """
    test_log_message = "Here's a test log message."
    dials.util.log.config_simple_cached()
    logger = logging.getLogger("dials")
    logger.info(test_log_message)
    (handler,) = logger.handlers
    return handler.records


@pytest.mark.xfail(
    "os.name == 'nt'",
    reason="https://github.com/dials/dials/issues/1650",
    raises=AttributeError,
)
def test_cached_log_records(caplog):

    # Generate some cached log messages in easy_mp child processes.
    results = multi_node_parallel_map(
        log_something,
        iterable=range(5),
        njobs=2,
        nproc=2,
        cluster_method="multiprocessing",
    )
    # Get all the resulting log records in a single flattened list.
    results = [record for records in results for record in records]

    results_batch = batch_multi_node_parallel_map(
        log_something,
        range(5),
        njobs=2,
        nproc=2,
        cluster_method="multiprocessing",
        callback=lambda _: None,
    )
    # Get all the resulting log records in a single flattened list.
    results_batch = [
        record for batch in results_batch for records in batch for record in records
    ]

    # Check that re-handling the messages in a logger in this process with a
    # threshold severity of WARNING results in no log records being emitted.
    with dials.util.log.LoggingContext("dials", logging.WARNING):
        dials.util.log.rehandle_cached_records(results)
        assert not caplog.records

        dials.util.log.rehandle_cached_records(results_batch)
        assert not caplog.records

    # Check that re-handling the messages in a logger in this process with a
    # threshold severity of INFO results in all the log records being emitted.
    with dials.util.log.LoggingContext("dials", logging.INFO):
        dials.util.log.rehandle_cached_records(results)
        assert caplog.records == results
        caplog.clear()

        dials.util.log.rehandle_cached_records(results_batch)
        assert caplog.records == results_batch
        caplog.clear()



from __future__ import annotations


class IntegrationAlgorithm:
    """A class to perform bayesian integration"""

    def __init__(self, **kwargs):
        pass

    def __call__(self, reflections, image_volume=None):
        """Process the reflections.

        :param reflections: The reflections to integrate
        :return: The list of integrated reflections
        """
        # Integrate and return the reflections
        if image_volume is None:
            intensity = reflections["shoebox"].bayesian_summation_intensity()
        else:
            raise RuntimeError("Image volume not supported at the moment")
        reflections["intensity.sum.value"] = intensity.observed_value()
        reflections["intensity.sum.variance"] = intensity.observed_variance()
        reflections["background.sum.value"] = intensity.background_value()
        reflections["background.sum.variance"] = intensity.background_variance()
        success = intensity.observed_success()
        reflections.set_flags(success, reflections.flags.integrated_sum)
        return success



"""
A datastructure for summing over groups of symmetry equivalent reflections.

This module defines a blocked datastructures for summing over groups of
symmetry equivalent reflections, as required for scaling.
"""

from __future__ import annotations

from typing import List, Optional, Union

import numpy as np
import pandas as pd
from orderedset import OrderedSet
from scipy.sparse import csc_matrix

from cctbx import crystal, miller, sgtbx, uctbx
from dxtbx import flumpy
from scitbx import sparse

from dials.algorithms.scaling.error_model.error_model import BasicErrorModel
from dials.array_family import flex


def map_indices_to_asu(miller_indices, space_group, anomalous=False):
    """Map the indices to the asymmetric unit."""
    crystal_symmetry = crystal.symmetry(space_group=space_group)
    miller_set = miller.set(
        crystal_symmetry=crystal_symmetry,
        indices=miller_indices,
        anomalous_flag=anomalous,
    )
    miller_set_in_asu = miller_set.map_to_asu()
    return miller_set_in_asu.indices()


def get_sorted_asu_indices(asu_indices, space_group, anomalous=False):
    """Return the sorted asu indices and the permutation selection."""
    crystal_symmetry = crystal.symmetry(space_group=space_group)
    miller_set_in_asu = miller.set(
        crystal_symmetry=crystal_symmetry, indices=asu_indices, anomalous_flag=anomalous
    )
    permuted = miller_set_in_asu.sort_permutation(by_value="packed_indices")
    sorted_asu_miller_index = asu_indices.select(permuted)
    return sorted_asu_miller_index, permuted


class IhTable:
    """
    A class to manage access to Ih_table blocks.

    The idea here is to split the data into blocks to allow parallelized
    computations, but within the blocks the data are sorted by dataset.
    In each block, there exists a block_selection_list which contains the indices
    for each dataset from the input reflection table.

    This class acts as a 'master' to setup the block structure and control access
    to the underlying blocks - only metadata is kept in this class after
    initialisation, the reflections etc are all contained in the blocks.

    Attributes:
        space_group: The space group for the dataset.
        Ih_table_blocks (list): A list of IhTableBlock instances. All symmetry
            equivalent reflections are recorded in the same block, to allow
            splitting of the dataset for parallelized computations.
        nblocks (int): The number of blocks in the Ih_table_blocks list.
        blocked_selection_list (list): A list of lists. bsl[i][j] is the selection
            list for block i, dataset j.
        n_datasets: The number of input reflection tables used to make the Ih_table.
        size: The number of reflections across all blocks
        asu_index_dict (dict): A dictionary, key: asu_miller_index, value tuple
            containing group_id and block_id (where group id is the group index
            within its block).
    """

    id_ = "IhTable"

    def __init__(
        self,
        reflection_tables: List[flex.reflection_table],
        space_group: sgtbx.space_group,
        indices_lists: Optional[List[flex.size_t]] = None,
        nblocks: int = 1,
        free_set_percentage: float = 0,
        free_set_offset: int = 0,
        additional_cols: Optional[List[str]] = None,
        anomalous: bool = False,
    ):
        """
        Distribute the input data into the required structure.

        The reflection data can be split into blocks, while the relevant
        metadata is also generated.

        A list of flex.size_t indices can be provided - this allows the
        reflection table data to maintain a reference to a dataset from which
        it was selected; these will be used when making the block selections.
        e.g selection = flex.bool([True, False, True])
            r_1 = r_master.select(selection)
            indices_list = selection.iselection() = flex.size_t([0, 2])
            then the block selection will contain 0 and 2 to refer back
            to the location of the data in r_master.
        """
        if indices_lists:
            assert len(indices_lists) == len(reflection_tables)
        self.anomalous = anomalous
        self._asu_index_dict = {}
        self._free_asu_index_dict = {}
        self.space_group = space_group
        self.n_work_blocks = nblocks
        self.n_datasets = len(reflection_tables)
        self.Ih_table_blocks = []
        self.blocked_selection_list = []
        self.properties_dict = {
            "n_unique_in_each_block": [],
            "n_reflections_in_each_block": {},
            "miller_index_boundaries": [],
        }
        self.free_set_percentage = free_set_percentage
        self._determine_required_block_structures(
            reflection_tables, free_set_percentage, free_set_offset
        )
        self._create_empty_Ih_table_blocks()
        for i, table in enumerate(reflection_tables):
            if indices_lists:
                self._add_dataset_to_blocks(
                    i, table, indices_lists[i], additional_cols=additional_cols
                )
            else:
                self._add_dataset_to_blocks(i, table, additional_cols=additional_cols)
        self.generate_block_selections()
        self.free_Ih_table = None
        if free_set_percentage > 0:
            self.extract_free_set()
            self.free_Ih_table = True
        self.calc_Ih()

    def update_data_in_blocks(
        self, data: flex.double, dataset_id: int, column: str = "intensity"
    ) -> None:
        """
        Update a given column across all blocks for a given dataset.

        Given an array of data (of the same size as the input reflection
        table) and the name of the column, use the internal data to split
        this up and set in individual blocks.
        """
        assert column in ["intensity", "variance", "inverse_scale_factor"]
        assert dataset_id in range(self.n_datasets)
        # split up data for blocks
        data = flumpy.to_numpy(data)
        for block in self.blocked_data_list:
            data_for_block = data[block.block_selections[dataset_id]]
            start = block.dataset_info[dataset_id]["start_index"]
            end = block.dataset_info[dataset_id]["end_index"]
            block.Ih_table.loc[
                np.arange(start=start, stop=end), column
            ] = data_for_block

    def get_block_selections_for_dataset(self, dataset: int) -> List[flex.size_t]:
        """Generate the block selection list for a given dataset."""
        assert dataset in range(self.n_datasets)
        if self.free_Ih_table:
            return [
                self.blocked_selection_list[i][dataset]
                for i in range(self.n_work_blocks + 1)
            ]
        return [
            self.blocked_selection_list[i][dataset] for i in range(self.n_work_blocks)
        ]

    @property
    def size(self) -> int:
        """Sum the sizes of all work blocks to give the total number of reflections."""
        if self.free_Ih_table:
            return sum(block.size for block in self.Ih_table_blocks[:-1])
        return sum(block.size for block in self.Ih_table_blocks)

    def generate_block_selections(self) -> None:
        """Generate and set an updated blocked_selection_list."""
        self.blocked_selection_list = [
            block.block_selections for block in self.Ih_table_blocks
        ]

    def update_weights(
        self, error_model: Optional[BasicErrorModel] = None, dataset_id: int = None
    ) -> None:
        """Update the error model in the blocks."""
        for block in self.Ih_table_blocks:
            block.update_weights(error_model, dataset_id)

    @property
    def blocked_data_list(self) -> List["IhTableBlock"]:
        """Return the list of IhTableBlock instances."""
        return self.Ih_table_blocks

    def set_derivatives(self, derivatives: sparse.matrix, block_id: int) -> None:
        """Set the derivatives matrix for a given block."""
        self.Ih_table_blocks[block_id].derivatives = derivatives

    def set_inverse_scale_factors(self, new_scales: np.array, block_id: int) -> None:
        """Set the inverse scale factors for a given block."""
        self.Ih_table_blocks[block_id].inverse_scale_factors = new_scales

    def calc_Ih(self, block_id: int = None) -> None:
        """Calculate the latest value of Ih, for a given block or for all blocks."""
        if block_id is not None:
            self.Ih_table_blocks[block_id].calc_Ih()
        else:
            for block in self.Ih_table_blocks:
                block.calc_Ih()

    def _determine_required_block_structures(
        self,
        reflection_tables: List[flex.reflection_table],
        free_set_percentage: float = 0,
        free_set_offset: int = 0,
    ) -> None:
        """
        Inspect the input to determine how to split into blocks.

        Extract the asu miller indices from the reflection table and
        add data to the asu_index_dict and properties dict.
        """
        joint_asu_indices = flex.miller_index()
        for table in reflection_tables:
            if "asu_miller_index" not in table:
                table["asu_miller_index"] = map_indices_to_asu(
                    table["miller_index"], self.space_group, self.anomalous
                )
            joint_asu_indices.extend(table["asu_miller_index"])
        sorted_joint_asu_indices, _ = get_sorted_asu_indices(
            joint_asu_indices, self.space_group, self.anomalous
        )
        if not sorted_joint_asu_indices:
            raise ValueError("No data found in input file(s)")

        asu_index_set = OrderedSet(sorted_joint_asu_indices)
        n_unique_groups = len(asu_index_set)
        n_free_groups = None
        interval_between_free_groups = None
        if free_set_percentage:
            n_free_groups = int(free_set_percentage * n_unique_groups / 100.0)
            n_work_groups = n_unique_groups - n_free_groups
            interval_between_free_groups = int(100 / free_set_percentage)
        else:
            n_work_groups = n_unique_groups
        self.n_work_blocks = min(self.n_work_blocks, n_work_groups)
        # first remove the free set groups
        if free_set_percentage:
            groups_for_free_set = np.full(n_unique_groups, False, dtype=bool)
            for_free = np.arange(
                0 + free_set_offset, n_unique_groups, interval_between_free_groups
            )
            groups_for_free_set[for_free] = True
            asu_index_set = np.array(list(asu_index_set))
            # work_asu_index_set = asu_index_set[~groups_for_free_set]
            free_asu_index_set = asu_index_set[groups_for_free_set]
        else:
            # work_asu_index_set = asu_index_set
            free_asu_index_set = None

        # also record how many unique groups go into each block
        group_boundaries = [
            int(i * n_unique_groups / self.n_work_blocks)
            for i in range(self.n_work_blocks)
        ]
        group_boundaries.append(n_unique_groups)

        next_boundary = group_boundaries[1]
        block_id = 0
        group_id_in_block_i = 0
        for i, index in enumerate(asu_index_set):
            if i == next_boundary:
                self.properties_dict["n_unique_in_each_block"].append(
                    group_id_in_block_i
                )
                self.properties_dict["miller_index_boundaries"].append(tuple(index))
                block_id += 1
                next_boundary = group_boundaries[block_id + 1]
                group_id_in_block_i = 0
            self._asu_index_dict[tuple(index)] = group_id_in_block_i
            group_id_in_block_i += 1
        # record the number in the last work block
        self.properties_dict["n_unique_in_each_block"].append(group_id_in_block_i)
        self.properties_dict["miller_index_boundaries"].append((10000, 10000, 10000))

        block_id += 1
        group_id_in_block_i = 0
        if free_asu_index_set is not None:
            for index in free_asu_index_set:
                # no boundaries as all go into the final block
                self._free_asu_index_dict[tuple(index)] = group_id_in_block_i
                group_id_in_block_i += 1
        # record the number in the free block
        self.properties_dict["n_unique_in_each_block"].append(group_id_in_block_i)
        self.properties_dict["miller_index_boundaries"].append((10000, 10000, 10000))
        # ^ to avoid bounds checking when in last group
        # need to know how many reflections will be in each block also

        block_id = 0
        idx_prev = 0
        boundary = self.properties_dict["miller_index_boundaries"][0]
        for i, index in enumerate(sorted_joint_asu_indices):
            if index == boundary:
                n_in_prev_group = i - idx_prev
                self.properties_dict["n_reflections_in_each_block"][
                    block_id
                ] = n_in_prev_group
                block_id += 1
                boundary = self.properties_dict["miller_index_boundaries"][block_id]
                idx_prev = i
        self.properties_dict["n_reflections_in_each_block"][block_id] = (
            len(sorted_joint_asu_indices) - idx_prev
        )

    def _create_empty_Ih_table_blocks(self) -> None:
        for n in range(self.n_work_blocks):
            n_refl_in_block = self.properties_dict["n_reflections_in_each_block"][n]
            n_groups_in_block = self.properties_dict["n_unique_in_each_block"][n]
            self.Ih_table_blocks.append(
                IhTableBlock(
                    n_groups=n_groups_in_block,
                    n_refl=n_refl_in_block,
                    n_datasets=self.n_datasets,
                )
            )

    def _add_dataset_to_blocks(
        self,
        dataset_id: int,
        reflections: flex.reflection_table,
        indices_array: Optional[flex.size_t] = None,
        additional_cols: Optional[List[str]] = None,
    ) -> None:
        sorted_asu_indices, perm = get_sorted_asu_indices(
            reflections["asu_miller_index"], self.space_group, self.anomalous
        )
        hkl = reflections["asu_miller_index"]
        df = pd.DataFrame()
        df["intensity"] = flumpy.to_numpy(reflections["intensity"])
        df["variance"] = flumpy.to_numpy(reflections["variance"])
        df["inverse_scale_factor"] = flumpy.to_numpy(
            reflections["inverse_scale_factor"]
        )
        if isinstance(additional_cols, list):
            for col in additional_cols:
                if col in reflections:
                    df[col] = flumpy.to_numpy(reflections[col])
        if indices_array:
            df["loc_indices"] = flumpy.to_numpy(indices_array)
        else:
            df["loc_indices"] = np.arange(df.shape[0], dtype=np.uint64)
        df = df.iloc[flumpy.to_numpy(perm)]
        hkl = hkl.select(perm)
        df["dataset_id"] = np.full(df.shape[0], dataset_id, dtype=np.uint64)
        # if data are sorted by asu_index, then up until boundary, should be in same
        # block (still need to read group_id though)
        # sort data, get group ids and block_ids
        group_ids = np.zeros(sorted_asu_indices.size(), dtype=np.uint64)
        boundary = self.properties_dict["miller_index_boundaries"][0]
        boundary_id = 0
        boundaries_for_this_datset = [0]  # use to slice
        # make this a c++ method for speed?
        prev = (0, 0, 0)
        group_id = -1
        for i, index in enumerate(sorted_asu_indices):
            if index != prev:
                while index >= boundary:
                    boundaries_for_this_datset.append(i)
                    boundary_id += 1
                    boundary = self.properties_dict["miller_index_boundaries"][
                        boundary_id
                    ]
                group_id = self._asu_index_dict[tuple(index)]
                prev = index
            group_ids[i] = group_id
        while len(boundaries_for_this_datset) < self.n_work_blocks + 1:
            # catch case where last boundaries aren't reached
            boundaries_for_this_datset.append(len(sorted_asu_indices))
        # so now have group ids as well for individual dataset
        if self.n_work_blocks == 1:
            self.Ih_table_blocks[0].add_data(dataset_id, group_ids, df, hkl)
        else:
            for i, val in enumerate(boundaries_for_this_datset[:-1]):
                start = val
                end = boundaries_for_this_datset[i + 1]
                self.Ih_table_blocks[i].add_data(
                    dataset_id, group_ids[start:end], df[start:end], hkl[start:end]
                )

    def extract_free_set(self) -> None:
        """Extract a free set from all blocks."""
        assert not self.free_Ih_table
        free_reflection_table = pd.DataFrame()
        free_indices = np.array([], dtype=int).reshape((0,))
        free_hkl = flex.miller_index([])
        # for each block, remove a fraction of the groups
        for j, block in enumerate(self.Ih_table_blocks):
            n_groups = block.n_groups
            groups_for_free_set = np.full(n_groups, False, dtype=bool)
            for_free = np.array(
                [
                    tuple(i) in self._free_asu_index_dict
                    for i in OrderedSet(block.asu_miller_index)
                ]
            )
            groups_for_free_set[for_free] = True
            free_block = block.select_on_groups(groups_for_free_set)
            free_reflection_table = pd.concat(
                [free_reflection_table, free_block.Ih_table]
            )
            free_hkl.extend(free_block.asu_miller_index)
            for sel in free_block.block_selections:
                free_indices = np.concatenate([free_indices, sel])
            self.Ih_table_blocks[j] = block.select_on_groups(~groups_for_free_set)
            # Now need to update dataset_info dict.
            removed_from_each_dataset = [
                np.count_nonzero(free_block.Ih_table["dataset_id"].to_numpy() == i)
                for i in range(0, block.n_datasets)
            ]
            n_removed = 0
            for i in range(0, self.Ih_table_blocks[j].n_datasets):
                self.Ih_table_blocks[j].dataset_info[i]["start_index"] -= n_removed
                n_removed += removed_from_each_dataset[i]
                self.Ih_table_blocks[j].dataset_info[i]["end_index"] -= n_removed
        self.blocked_selection_list = [
            block.block_selections for block in self.Ih_table_blocks
        ]
        # now split by dataset and use to instantiate another Ih_table
        datasets = set(free_reflection_table["dataset_id"])
        tables = []
        indices_lists = []
        n_refl = 0
        for id_ in datasets:
            dataset_sel = free_reflection_table["dataset_id"].to_numpy() == id_
            n_refl += np.count_nonzero(dataset_sel)
            tables.append(free_reflection_table[dataset_sel])
            indices_lists.append(free_indices[dataset_sel])
        free_block = IhTableBlock(
            n_groups=len(set(free_hkl)), n_refl=n_refl, n_datasets=len(datasets)
        )
        group_ids = np.array(
            [self._free_asu_index_dict[tuple(index)] for index in free_hkl],
            dtype=np.uint64,
        )
        for id_, t in zip(datasets, tables):
            dataset_sel = free_reflection_table["dataset_id"].to_numpy() == id_
            group_id_this = group_ids[dataset_sel]
            hkl_this = free_hkl.select(flumpy.from_numpy(dataset_sel))
            free_block.add_data(id_, group_id_this, t, hkl_this)

        self.Ih_table_blocks.append(free_block)
        self.blocked_selection_list.append(free_block.block_selections)

    def as_miller_array(
        self, unit_cell: uctbx.unit_cell, return_free_set_data: bool = False
    ) -> miller.array:
        """Get a scaled miller array from the Ih_table and an experiment."""
        blocked_data_list = self.blocked_data_list
        joint_table = flex.reflection_table([])
        if self.free_Ih_table:
            if return_free_set_data:
                blocked_data_list = [blocked_data_list[-1]]
            else:
                blocked_data_list = blocked_data_list[:-1]
        if len(blocked_data_list) > 1:
            for block in blocked_data_list:
                joint_table.extend(block.as_reflection_table())
        else:
            joint_table = blocked_data_list[0].as_reflection_table()
        # Filter out negative scale factors to avoid merging statistics errors.
        return _reflection_table_to_iobs(joint_table, unit_cell, self.space_group)


class TargetAsuDictCache(object):

    instances = {}

    def __new__(cls, target_Ih_table):
        id_ = id(target_Ih_table)
        if id_ not in cls.instances:
            cls.instances[id_] = dict(
                zip(
                    target_Ih_table.blocked_data_list[0].asu_miller_index,
                    target_Ih_table.blocked_data_list[0].Ih_values,
                )
            )
        return cls.instances[id_]


class IhTableBlock:
    """
    A datastructure for efficient summations over symmetry equivalent reflections.

    This contains a reflection table, sorted by dataset, called the Ih_table,
    a h_index_matrix (sparse) for efficiently calculating sums over symmetry
    equivalent reflections as well as 'block_selections' which relate the order
    of the data to the initial reflection tables used to initialise the (master)
    IhTable.

    Attributes:
        Ih_table: A reflection table, containing I, g, w, var, Ih,
            asu_miller_index, loc_indices and dataset_id.
        block_selections: A list of flex.size_t arrays of indices, that can be
            used to select and reorder data from the input reflection tables to
            match the order in the Ih_table.
        h_index_matrix: A sparse matrix used to sum over groups of equivalent
            reflections by multiplication. Sum_h I = I * h_index_matrix. The
            dimension is n_refl by n_groups; each row has a single nonzero
            entry with a value of 1.
        h_expand_matrix: The transpose of the h_index_matrix, used to expand an
            array of values for symmetry groups into an array of size n_refl.
        derivatives: A matrix of derivatives of the reflections wrt the model
            parameters.
    """

    def __init__(self, n_groups: int, n_refl: int, n_datasets: int = 1):
        """Create empty datastructures to which data can later be added."""
        self.Ih_table = pd.DataFrame()
        self.block_selections = [None] * n_datasets
        self.h_index_matrix = sparse.matrix(n_refl, n_groups)
        self._setup_info = {"next_row": 0, "next_dataset": 0, "setup_complete": False}
        self.dataset_info = {}
        self.n_datasets = n_datasets
        self.h_expand_matrix = None
        self.derivatives = None
        self.binner = None
        self._csc_rows = np.array([], dtype=np.uint64).reshape((0,))
        self._csc_cols = np.array([], dtype=np.uint64).reshape((0,))
        self._csc_h_index_matrix = None
        self._csc_h_expand_matrix = None
        self._hkl = flex.miller_index([])

    def add_data(
        self,
        dataset_id: int,
        group_ids: np.array,
        reflections: pd.DataFrame,
        hkl: flex.miller_index,
    ) -> None:
        """
        Add data to all blocks for a given dataset.

        Add data to the Ih_table, write data to the h_index_matrix and
        add the loc indices to the block_selections list.
        """
        assert not self._setup_info[
            "setup_complete"
        ], """
No further data can be added to the IhTableBlock as setup marked complete."""
        assert (
            self._setup_info["next_row"] + len(group_ids) <= self.h_index_matrix.n_rows
        ), """
Not enough space left to add this data, please check for correct block initialisation."""
        assert (
            dataset_id == self._setup_info["next_dataset"]
        ), """
Datasets must be added in correct order: expected: {}, this dataset: {}""".format(
            self._setup_info["next_dataset"],
            dataset_id,
        )
        for i, id_ in enumerate(group_ids):
            rowidx = i + self._setup_info["next_row"]
            self.h_index_matrix[rowidx, int(id_)] = 1.0
        cols = group_ids
        rows = np.arange(
            start=self._setup_info["next_row"],
            stop=self._setup_info["next_row"] + group_ids.size,
            dtype=np.uint64,
        )

        self._csc_cols = np.concatenate([self._csc_cols, cols])
        self._csc_rows = np.concatenate([self._csc_rows, rows])
        self._hkl.extend(hkl)

        self.dataset_info[dataset_id] = {"start_index": self._setup_info["next_row"]}
        self._setup_info["next_row"] += len(group_ids)
        self._setup_info["next_dataset"] += 1
        self.dataset_info[dataset_id]["end_index"] = self._setup_info["next_row"]
        self.Ih_table = pd.concat([self.Ih_table, reflections], ignore_index=True)
        if "loc_indices" in reflections:
            self.block_selections[dataset_id] = reflections["loc_indices"].to_numpy()
        else:
            self.block_selections[dataset_id] = np.arange(
                reflections.shape[0], dtype=np.uint64
            )
        if self._setup_info["next_dataset"] == len(self.block_selections):
            self._complete_setup()

    def _complete_setup(self) -> None:
        """Finish the setup of the Ih_table once all data has been added."""
        self.h_index_matrix.compact()
        assert (
            self._setup_info["next_row"] == self.h_index_matrix.n_rows
        ), """
Not all rows of h_index_matrix appear to be filled in IhTableBlock setup."""
        self.h_expand_matrix = self.h_index_matrix.transpose()
        data = np.full(self._csc_cols.size, 1.0)
        self._csc_h_index_matrix = csc_matrix((data, (self._csc_rows, self._csc_cols)))
        self._csc_h_expand_matrix = self._csc_h_index_matrix.transpose()
        self.weights = 1.0 / self.variances
        self._setup_info["setup_complete"] = True

    def group_multiplicities(self, output: str = "per_group") -> np.array:
        """Return the multiplicities of the symmetry groups."""
        return self.sum_in_groups(np.full(self.size, 1.0), output=output)

    def select(self, sel: np.array) -> "IhTableBlock":
        """Select a subset of the data, returning a new IhTableBlock object."""
        Ih_table = self.Ih_table[sel]
        Ih_table.reset_index(drop=True, inplace=True)
        h_idx_sel = self.h_expand_matrix.select_columns(
            flumpy.from_numpy(sel).iselection()
        )
        reduced_h_idx = h_idx_sel.transpose()
        unity = flex.double(int(Ih_table.size), 1.0)
        nz_col_sel = (unity * reduced_h_idx) > 0
        h_index_matrix = reduced_h_idx.select_columns(nz_col_sel.iselection())
        h_expand = h_index_matrix.transpose()
        csc_h_idx_sel = self._csc_h_expand_matrix[:, sel]
        csc_h_index_matrix = csc_h_idx_sel.transpose()[:, flumpy.to_numpy(nz_col_sel)]
        csc_h_expand_matrix = csc_h_index_matrix.transpose()
        newtable = IhTableBlock(n_groups=0, n_refl=0, n_datasets=self.n_datasets)
        newtable.Ih_table = Ih_table
        newtable._hkl = self._hkl.select(flumpy.from_numpy(sel))
        newtable.h_expand_matrix = h_expand
        newtable.h_index_matrix = h_index_matrix
        newtable._csc_h_index_matrix = csc_h_index_matrix
        newtable._csc_h_expand_matrix = csc_h_expand_matrix
        newtable.block_selections = []
        offset = 0
        for i in range(newtable.n_datasets):
            newtable.dataset_info[i] = {"start_index": offset}
            block_sel_i = self.block_selections[i]
            n_in_dataset_i = len(block_sel_i)
            newtable.block_selections.append(
                block_sel_i[sel[offset : offset + n_in_dataset_i]]
            )
            offset += n_in_dataset_i
            newtable.dataset_info[i]["end_index"] = offset
        return newtable

    def select_on_groups(self, sel: np.array) -> "IhTableBlock":
        """Select a subset of the unique groups, returning a new IhTableBlock."""
        reduced_h_idx = self._csc_h_index_matrix[:, sel]
        unity = np.full(reduced_h_idx.shape[1], 1.0)
        nz_row_sel = (unity * reduced_h_idx.transpose()) > 0
        return self.select(nz_row_sel)

    def calc_Ih(self) -> None:
        """Calculate the current best estimate for Ih for each reflection group."""
        scale_factors = self.inverse_scale_factors
        sumgsq = self.sum_in_groups(np.square(scale_factors) * self.weights)
        sumgI = self.sum_in_groups(scale_factors * self.intensities * self.weights)
        Ih = sumgI / sumgsq
        self.Ih_table.loc[:, "Ih_values"] = Ih @ self._csc_h_expand_matrix

    def update_weights(
        self,
        error_model: Optional[BasicErrorModel] = None,
        dataset_id: Optional[int] = None,
    ) -> None:
        """Update the scaling weights based on an error model."""
        if error_model:
            if dataset_id is not None:  # note the first dataset has an id of 0
                sel = self.Ih_table["dataset_id"].to_numpy() == dataset_id
                sigmaprimesq = error_model.update_variances(
                    self.variances[sel], self.intensities[sel]
                )
                self.Ih_table.loc[sel, "weights"] = 1.0 / sigmaprimesq
            else:
                sigmaprimesq = error_model.update_variances(
                    self.variances, self.intensities
                )
                self.Ih_table.loc[:, "weights"] = 1.0 / sigmaprimesq
        else:
            if dataset_id is not None:  # note the first dataset has an id of 0
                sel = self.Ih_table["dataset_id"].to_numpy() == dataset_id
                self.Ih_table.loc[sel, "weights"] = 1.0 / self.variances[sel]
            else:
                self.Ih_table.loc[:, "weights"] = 1.0 / self.variances

    def calc_nh(self) -> np.array:
        """Calculate the number of refls in the group to which the reflection belongs.

        This is a vector of length n_refl."""
        return self.sum_in_groups(np.full(self.size, 1.0), output="per_refl")

    def match_Ih_values_to_target(self, target_Ih_table: IhTable) -> None:
        """
        Use an Ih_table as a target to set Ih values in this table.

        Given an Ih table as a target, the common reflections across the tables
        are determined and the Ih_values are set to those of the target. If no
        matching reflection is found, then the values are removed from the table.
        """
        assert target_Ih_table.n_work_blocks == 1
        target_asu_Ih_dict = TargetAsuDictCache(target_Ih_table)
        new_Ih_values = np.zeros(self.size, dtype=float)
        location_in_unscaled_array = 0
        sorted_asu_indices, permuted = get_sorted_asu_indices(
            self.asu_miller_index,
            target_Ih_table.space_group,
            anomalous=target_Ih_table.anomalous,
        )
        for j, miller_idx in enumerate(OrderedSet(sorted_asu_indices)):
            n_in_group = self._csc_h_index_matrix.getcol(j).count_nonzero()
            if miller_idx in target_asu_Ih_dict:
                i = location_in_unscaled_array
                new_Ih_values[np.arange(i, i + n_in_group, dtype=np.uint64)] = np.full(
                    n_in_group, target_asu_Ih_dict[miller_idx]
                )
            location_in_unscaled_array += n_in_group
        self.Ih_table.loc[flumpy.to_numpy(permuted), "Ih_values"] = new_Ih_values
        sel = self.Ih_values != 0.0
        new_table = self.select(sel)
        # now set attributes to update object
        self.Ih_table = new_table.Ih_table
        self.h_index_matrix = new_table.h_index_matrix
        self.h_expand_matrix = new_table.h_expand_matrix
        self.block_selections = new_table.block_selections
        self._csc_h_expand_matrix = new_table._csc_h_expand_matrix
        self._csc_h_index_matrix = new_table._csc_h_index_matrix

    @property
    def inverse_scale_factors(self) -> np.array:
        """The inverse scale factors of the reflections."""
        return self.Ih_table["inverse_scale_factor"].to_numpy()

    @inverse_scale_factors.setter
    def inverse_scale_factors(self, new_scales: np.array) -> None:
        if new_scales.size != self.size:
            assert 0, """attempting to set a new set of scale factors of different
      length than previous assignment: was {}, attempting {}""".format(
                self.inverse_scale_factors.size,
                new_scales.size,
            )
        else:
            self.Ih_table.loc[:, "inverse_scale_factor"] = new_scales

    @property
    def variances(self) -> np.array:
        """The variances of the reflections."""
        return self.Ih_table["variance"].to_numpy()

    @variances.setter
    def variances(self, new_variances: np.array) -> None:
        assert new_variances.size == self.size
        self.Ih_table.loc[:, "variance"] = new_variances

    @property
    def intensities(self) -> np.array:
        """The unscaled reflection intensities."""
        return self.Ih_table["intensity"].to_numpy()

    @intensities.setter
    def intensities(self, new_intensities):
        assert new_intensities.size == self.size
        self.Ih_table.loc[:, "intensity"] = new_intensities

    @property
    def Ih_values(self) -> np.array:
        """The bset-estimated intensities of symmetry equivalent reflections."""
        return self.Ih_table["Ih_values"].to_numpy()

    @property
    def weights(self) -> np.array:
        """The weights that will be used in scaling."""
        return self.Ih_table["weights"].to_numpy()

    @weights.setter
    def weights(self, new_weights):
        if new_weights.size != self.size:
            assert 0, """attempting to set a new set of weights of different
      length than previous assignment: was {}, attempting {}""".format(
                self.size,
                new_weights.size,
            )
        self.Ih_table.loc[:, "weights"] = new_weights

    @property
    def size(self) -> int:
        """Return the length of the stored Ih_table (a reflection table)."""
        return self.Ih_table.shape[0]

    @property
    def n_groups(self) -> int:
        """Return the length of the stored Ih_table (a reflection table)."""
        return self._csc_h_index_matrix.shape[1]

    @property
    def asu_miller_index(self) -> flex.miller_index:
        """Return the miller indices in the asymmetric unit."""
        return self._hkl

    def setup_binner(
        self,
        unit_cell: uctbx.unit_cell,
        space_group: sgtbx.space_group,
        n_resolution_bins: int,
    ) -> None:
        """Create a binner for the reflections contained in the table."""
        ma = _reflection_table_to_iobs(
            self.as_reflection_table(), unit_cell, space_group
        )
        # need d star sq step
        d_star_sq = ma.d_star_sq().data()
        d_star_sq_min = flex.min(d_star_sq)
        d_star_sq_max = flex.max(d_star_sq)
        span = d_star_sq_max - d_star_sq_min
        relative_tolerance = 1e-6
        d_star_sq_max += span * relative_tolerance
        d_star_sq_min -= span * relative_tolerance
        # Avoid a zero-size step that would otherwise anger the d_star_sq_step binner.
        step = max((d_star_sq_max - d_star_sq_min) / n_resolution_bins, 0.004)

        self.binner = ma.setup_binner_d_star_sq_step(
            auto_binning=False,
            d_max=uctbx.d_star_sq_as_d(d_star_sq_max),
            d_min=uctbx.d_star_sq_as_d(d_star_sq_min),
            d_star_sq_step=step,
        )

    def sum_in_groups(
        self, array: Union[csc_matrix, np.array], output: str = "per_group"
    ) -> np.array:
        """
        Sums an array object over the symmetry equivalent groups.
        The array's final dimension must equal the size of the Ih_table.
        """
        if output == "per_group":
            return array @ self._csc_h_index_matrix
        elif output == "per_refl":  # return the summed quantity per reflection
            return (array @ self._csc_h_index_matrix) @ self._csc_h_expand_matrix
        else:
            raise ValueError(
                f"""Bad value for output= parameter
(value={output}, allowed values: per_group, per_refl)"""
            )

    def as_reflection_table(self) -> flex.reflection_table:
        """Return the data in flex reflection table format"""
        table = flex.reflection_table()
        table["asu_miller_index"] = self.asu_miller_index
        for k, v in self.Ih_table.iteritems():
            table[k] = flumpy.from_numpy(v.to_numpy())
        return table


def _reflection_table_to_iobs(
    table: flex.reflection_table,
    unit_cell: uctbx.unit_cell,
    space_group: sgtbx.space_group,
) -> miller.array:
    miller_set = miller.set(
        crystal_symmetry=crystal.symmetry(
            unit_cell=unit_cell,
            space_group=space_group,
            assert_is_compatible_unit_cell=False,
        ),
        indices=table["asu_miller_index"],
        anomalous_flag=False,
    )
    i_obs = miller.array(
        miller_set, data=table["intensity"] / table["inverse_scale_factor"]
    )
    i_obs.set_observation_type_xray_intensity()
    i_obs.set_sigmas(flex.sqrt(table["variance"]) / table["inverse_scale_factor"])
    i_obs.set_info(miller.array_info(source="DIALS", source_type="reflection_tables"))
    return i_obs



from __future__ import annotations

import random
import sys
import time
from math import pi

import numpy as np
import pytest

import dxtbx.flumpy as flumpy
from dxtbx.model.experiment_list import Experiment, ExperimentList
from scitbx.array_family import flex

from dials.algorithms.refinement.parameterisation.scan_varying_beam_parameters import (
    ScanVaryingBeamParameterisation,
)
from dials.algorithms.refinement.parameterisation.scan_varying_crystal_parameters import (
    ScanVaryingCrystalOrientationParameterisation,
    ScanVaryingCrystalUnitCellParameterisation,
)
from dials.algorithms.refinement.parameterisation.scan_varying_detector_parameters import (
    ScanVaryingDetectorParameterisationSinglePanel,
)
from dials.algorithms.refinement.parameterisation.scan_varying_goniometer_parameters import (
    ScanVaryingGoniometerParameterisation,
)
from dials.algorithms.refinement.parameterisation.scan_varying_prediction_parameters import (
    ScanVaryingPredictionParameterisation,
    SparseFlex,
)
from dials.algorithms.refinement.prediction.managed_predictors import (
    ScansExperimentsPredictor,
    ScansRayPredictor,
)
from dials_refinement_helpers_ext import (
    build_reconstitute_derivatives_mat3,
    build_reconstitute_derivatives_vec3,
    intersection_i_seqs_unsorted,
)


class _Test:
    def create_models(self, cmdline_overrides=None):
        from dxtbx.model import ScanFactory
        from libtbx.phil import parse

        from . import geometry_phil
        from .setup_geometry import Extract

        if cmdline_overrides is None:
            cmdline_overrides = []
        overrides = """geometry.parameters.crystal.a.length.range = 10 50
geometry.parameters.crystal.b.length.range = 10 50
geometry.parameters.crystal.c.length.range = 10 50"""

        master_phil = parse(geometry_phil)

        # Extract models
        models = Extract(master_phil, overrides, cmdline_args=cmdline_overrides)
        self.detector = models.detector
        self.goniometer = models.goniometer
        self.crystal = models.crystal
        self.beam = models.beam

        # Make a scan of 1-20 * 0.5 deg images
        sf = ScanFactory()
        self.scan = sf.make_scan((1, 20), 0.5, (0, 0.5), list(range(20)))

        # Generate an ExperimentList
        self.experiments = ExperimentList()
        self.experiments.append(
            Experiment(
                beam=self.beam,
                detector=self.detector,
                goniometer=self.goniometer,
                scan=self.scan,
                crystal=self.crystal,
                imageset=None,
            )
        )

        # Create a reflection predictor for the experiments
        self.ref_predictor = ScansExperimentsPredictor(self.experiments)

        # Create scan-varying parameterisations of these models, with 3 samples
        self.det_param = ScanVaryingDetectorParameterisationSinglePanel(
            self.detector, self.scan.get_array_range(), 3
        )
        self.s0_param = ScanVaryingBeamParameterisation(
            self.beam, self.scan.get_array_range(), 3, self.goniometer
        )
        self.xlo_param = ScanVaryingCrystalOrientationParameterisation(
            self.crystal, self.scan.get_array_range(), 3
        )
        self.xluc_param = ScanVaryingCrystalUnitCellParameterisation(
            self.crystal, self.scan.get_array_range(), 3
        )
        self.gon_param = ScanVaryingGoniometerParameterisation(
            self.goniometer, self.scan.get_array_range(), 3, self.beam
        )

    def generate_reflections(self):
        from cctbx.sgtbx import space_group, space_group_symbols

        from dials.algorithms.spot_prediction import IndexGenerator, ray_intersection

        sequence_range = self.scan.get_oscillation_range(deg=False)
        resolution = 2.0
        index_generator = IndexGenerator(
            self.crystal.get_unit_cell(),
            space_group(space_group_symbols(1).hall()).type(),
            resolution,
        )
        indices = index_generator.to_array()

        # Predict rays within the sequence range
        ray_predictor = ScansRayPredictor(self.experiments, sequence_range)
        obs_refs = ray_predictor(indices)

        # Take only those rays that intersect the detector
        intersects = ray_intersection(self.detector, obs_refs)
        obs_refs = obs_refs.select(intersects)

        # Re-predict using the Experiments predictor for all these reflections. The
        # result is the same, but we gain also the flags and xyzcal.px columns
        obs_refs["id"] = flex.int(len(obs_refs), 0)
        obs_refs = self.ref_predictor(obs_refs)

        # Set 'observed' centroids from the predicted ones
        obs_refs["xyzobs.mm.value"] = obs_refs["xyzcal.mm"]

        # Invent some variances for the centroid positions of the simulated data
        im_width = 0.1 * pi / 180.0
        px_size = self.detector[0].get_pixel_size()
        var_x = flex.double(len(obs_refs), (px_size[0] / 2.0) ** 2)
        var_y = flex.double(len(obs_refs), (px_size[1] / 2.0) ** 2)
        var_phi = flex.double(len(obs_refs), (im_width / 2.0) ** 2)
        obs_refs["xyzobs.mm.variance"] = flex.vec3_double(var_x, var_y, var_phi)

        # set the flex random seed to an 'uninteresting' number
        flex.set_random_seed(12407)

        # take 10 random reflections for speed
        reflections = obs_refs.select(flex.random_selection(len(obs_refs), 10))

        # use a BlockCalculator to calculate the blocks per image
        from dials.algorithms.refinement.reflection_manager import BlockCalculator

        block_calculator = BlockCalculator(self.experiments, reflections)
        reflections = block_calculator.per_image()

        return reflections


def test(cmdline_overrides=[]):
    tc = _Test()
    tc.create_models(cmdline_overrides)
    reflections = tc.generate_reflections()

    # use a ReflectionManager to exclude reflections too close to the spindle,
    # plus set the frame numbers
    from dials.algorithms.refinement.reflection_manager import ReflectionManager

    refman = ReflectionManager(reflections, tc.experiments, outlier_detector=None)
    refman.finalise()

    # create prediction parameterisation of the requested type
    pred_param = ScanVaryingPredictionParameterisation(
        tc.experiments,
        [tc.det_param],
        [tc.s0_param],
        [tc.xlo_param],
        [tc.xluc_param],
        [tc.gon_param],
    )

    # keep only those reflections that pass inclusion criteria and have predictions
    reflections = refman.get_matches()

    # get analytical gradients
    pred_param.compose(reflections)
    an_grads = pred_param.get_gradients(reflections)

    # get finite difference gradients
    p_vals = pred_param.get_param_vals()
    p_names = pred_param.get_param_names()
    deltas = [1.0e-7] * len(p_vals)

    for i, delta in enumerate(deltas):
        val = p_vals[i]

        p_vals[i] -= delta / 2.0
        pred_param.set_param_vals(p_vals)
        pred_param.compose(reflections)

        tc.ref_predictor(reflections)

        rev_state = reflections["xyzcal.mm"].deep_copy()

        p_vals[i] += delta
        pred_param.set_param_vals(p_vals)
        pred_param.compose(reflections)

        tc.ref_predictor(reflections)

        fwd_state = reflections["xyzcal.mm"].deep_copy()
        p_vals[i] = val

        fd = fwd_state - rev_state
        x_grads, y_grads, phi_grads = fd.parts()
        x_grads /= delta
        y_grads /= delta
        phi_grads /= delta

        try:
            for (a, b) in zip(x_grads, an_grads[i]["dX_dp"]):
                assert a == pytest.approx(b, abs=1e-5)
            for (a, b) in zip(y_grads, an_grads[i]["dY_dp"]):
                assert a == pytest.approx(b, abs=1e-5)
            for (a, b) in zip(phi_grads, an_grads[i]["dphi_dp"]):
                assert a == pytest.approx(b, abs=1e-5)
        except AssertionError:
            print(f"Failure for {p_names[i]}")
            raise

    # return to the initial state
    pred_param.set_param_vals(p_vals)
    pred_param.compose(reflections)


def test_SparseFlex_scalars():

    size = 100

    # Make a dense double array with 50% explicit zeroes
    arr = flex.random_double(size)
    indices = flex.random_selection(size, int(size / 2))
    elements = arr.select(indices)
    arr *= 0.0
    arr.set_selected(indices, elements)

    # Make the equivalent SparseFlex
    sf_arr = SparseFlex(size, elements, indices)

    # Test multiplication of SparseFlex[double] by scalar
    sf2 = sf_arr * 2.0  # __mul__
    for a, b in zip(sf2.as_dense_vector(), arr * 2.0):
        assert a == b

    sf2 = 2.0 * sf_arr  # __rmul__
    for a, b in zip(sf2.as_dense_vector(), 2.0 * arr):
        assert a == b

    # Test division of SparseFlex[double] by scalar
    sf_half = sf_arr / 2.0
    for a, b in zip(sf_half.as_dense_vector(), arr / 2.0):
        assert a == b

    # Test addition of two SparseFlex[double]s
    sf2 = sf_arr + sf_arr
    for a, b in zip(sf2.as_dense_vector(), arr + arr):
        assert a == b

    # Test subtraction between two SparseFlex[double]s
    sf3 = sf2 - sf_arr
    for a, b in zip(sf3.as_dense_vector(), arr):
        assert a == b


def test_SparseFlex_matrix_and_vector_arithmetic():

    size = 100

    # Make a dense vec3 array with 50% explicit zeroes
    vec = flex.vec3_double(
        flex.random_double(size), flex.random_double(size), flex.random_double(size)
    )
    indices = flex.random_selection(size, int(size / 2))
    elements = vec.select(indices)
    vec *= 0.0
    vec.set_selected(indices, elements)

    # Make the equivalent SparseFlex[vec3]
    sf_vec = SparseFlex(size, elements, indices)

    # Make a dense mat3 array with the same 50% explicit zeroes
    mat = flex.mat3_double(
        (flex.random_double_r3_rotation_matrix() for i in range(size))
    )
    elements = mat.select(indices)
    mat *= 0.0
    mat.set_selected(indices, elements)

    # Make the equivalent SparseFlex[mat3]
    sf_mat = SparseFlex(size, elements, indices)

    # Test multiplication of SparseFlex[vec3] by scalar
    sf2 = sf_vec * 2.0  # __mul__
    for a, b in zip(sf2.as_dense_vector(), vec * 2.0):
        assert a == b

    sf2 = 2.0 * sf_vec  # __rmul__
    for a, b in zip(sf2.as_dense_vector(), vec * 2.0):
        assert a == b

    # Test multiplication of SparseFlex[mat3] by scalar. Only __mul__, because
    # mat3_double does not have __rmul__ defined
    sf2 = sf_mat * 2.0
    for a, b in zip(sf2.as_dense_vector(), mat * 2.0):
        assert a == b

    # Test matrix multiplication: SparseFlex[mat3] * flex.vec3_double. Use a
    # new vector which does not have explicit zero elements
    vec2 = flex.vec3_double(
        flex.random_double(size), flex.random_double(size), flex.random_double(size)
    )
    sf_rot = sf_mat * vec2
    for a, b in zip(sf_rot.as_dense_vector(), mat * vec2):
        assert a == b

    # Test matrix multiplication: SparseFlex[mat3] * SparseFlex[vec3]
    sf_rot = sf_mat * sf_vec
    for a, b in zip(sf_rot.as_dense_vector(), mat * vec):
        assert a == b

    # Test matrix multiplication: SparseFlex[mat3] * flex.mat3_double. Use a
    # new matrix which does not have explicit zero elements
    mat2 = flex.mat3_double(
        (flex.random_double_r3_rotation_matrix() for i in range(size))
    )
    sf_mat2 = sf_mat * mat2
    for a, b in zip(sf_mat2.as_dense_vector(), mat * mat2):
        assert a == b

    # Test matrix multiplication SparseFlex[mat3] * SparseFlex[mat3]
    sf_mat2 = sf_mat * sf_mat
    for a, b in zip(sf_mat2.as_dense_vector(), mat * mat):
        assert a == b

    # Test addition of two SparseFlex[vec3]s
    sf2 = sf_vec + sf_vec
    for a, b in zip(sf2.as_dense_vector(), vec + vec):
        assert a == b

    # Test subtraction between two SparseFlex[vec3]s
    sf3 = sf2 - sf_vec
    for a, b in zip(sf3.as_dense_vector(), vec):
        assert a == b


def test_SparseFlex_vec3_only_methods():

    size = 100

    # Make a dense vec3 array with 50% explicit zeroes
    vec = flex.vec3_double(
        flex.random_double(size), flex.random_double(size), flex.random_double(size)
    )
    indices = flex.random_selection(size, int(size / 2))
    elements = vec.select(indices)
    vec *= 0.0
    vec.set_selected(indices, elements)

    # Make the equivalent SparseFlex[vec3]
    sf_vec = SparseFlex(size, elements, indices)

    # Test dot product of SparseFlex[vec3] and flex.vec3_double. Use a
    # new vector which does not have explicit zero elements
    vec2 = flex.vec3_double(
        flex.random_double(size), flex.random_double(size), flex.random_double(size)
    )
    dotprod = sf_vec.dot(vec2)
    for a, b in zip(dotprod.as_dense_vector(), vec.dot(vec2)):
        assert a == b

    # Test dot product of SparseFlex[vec3] and SparseFlex[vec3]
    dotprod = sf_vec.dot(sf_vec)
    for a, b in zip(dotprod.as_dense_vector(), vec.dot(vec)):
        assert a == b

    # Test rotate_around_origin
    directions = flex.vec3_double(
        flex.random_double(size), flex.random_double(size), flex.random_double(size)
    )
    angles = flex.random_double(size) * pi
    sf_rot = sf_vec.rotate_around_origin(directions, angles)
    for a, b in zip(
        sf_rot.as_dense_vector(), vec.rotate_around_origin(directions, angles)
    ):
        assert a == b

    # Test parts
    sf_x, sf_y, sf_z = sf_vec.parts()
    x, y, z = vec.parts()
    for a, b in zip(sf_x.as_dense_vector(), x):
        assert a == b
    for a, b in zip(sf_y.as_dense_vector(), y):
        assert a == b
    for a, b in zip(sf_z.as_dense_vector(), z):
        assert a == b


def test_SparseFlex_select():

    size = 100

    # Make a dense double array with 50% explicit zeroes
    arr = flex.random_double(size)
    indices = flex.random_selection(size, int(size / 2))
    elements = arr.select(indices)
    arr *= 0.0
    arr.set_selected(indices, elements)

    # Make the equivalent SparseFlex
    sf_arr = SparseFlex(size, elements, indices)

    # Make a new random selection
    isel = flex.random_selection(size, int(size / 2))

    # Select
    sf2 = sf_arr.select(isel)
    arr2 = arr.select(isel)

    # Check
    for a, b in zip(sf2.as_dense_vector(), arr2):
        assert a == b


@pytest.mark.parametrize(
    "random_order", [False, pytest.param(True, marks=pytest.mark.xfail)]
)
def test_SparseFlex_select_intersection(random_order):
    """SparseFlex.select requires the intersection of two size_t arrays, A and
    B. We view this as the selection of elements in A according to their
    presence in B. We have three approaches for this: a pure Python method,
    a C++ equivalent and a Numpy method. The Numpy method is fast, but fails
    when the input arrays are not sorted. We cannot assume this to be the case
    in practice, so have to reject that method for use in SparseFlex.select.

    With changes to the C++ version, this is now about 8 times faster than the
    pure Python version and 4 times faster than the NumPy version."""

    # Two random selections
    size = 1000
    sel1 = flex.random_selection(size, int(size / 2))
    if random_order:
        perm = flex.random_permutation(int(size / 2))
        sel1 = sel1.select(perm)
    sel2 = flex.random_selection(size, int(size / 2))
    if random_order:
        perm = flex.random_permutation(int(size / 2))
        sel2 = sel2.select(perm)

    # Pure Python version
    start = time.perf_counter_ns()
    index_a = flex.size_t(0)
    index_b = flex.size_t(0)
    lookup = {}
    for i_a, val in enumerate(sel1):
        lookup[val] = i_a
    for i_b, val in enumerate(sel2):
        i_a = lookup.get(val)
        if i_a is not None:
            index_a.append(i_a)
            index_b.append(i_b)
    end = time.perf_counter_ns()
    wc_time_py = end - start

    # C++ version
    start = time.perf_counter_ns()
    index_a_cpp, index_b_cpp = intersection_i_seqs_unsorted(sel1, sel2)
    end = time.perf_counter_ns()
    wc_time_cpp = end - start

    # Numpy version
    start = time.perf_counter_ns()
    _, index_a_np, index_b_np = np.intersect1d(
        flumpy.to_numpy(sel1),
        flumpy.to_numpy(sel2),
        assume_unique=True,
        return_indices=True,
    )
    index_a_np = flumpy.from_numpy(index_a_np)
    index_b_np = flumpy.from_numpy(index_b_np)
    end = time.perf_counter_ns()
    wc_time_numpy = end - start

    print(f"Timing info Python: {wc_time_py} C++ {wc_time_cpp} Numpy: {wc_time_numpy}")

    # Check results are equal: Python to C++
    for (
        a,
        b,
    ) in zip(index_a_cpp, index_a):
        assert a == b
    for (
        a,
        b,
    ) in zip(index_b_cpp, index_b):
        assert a == b

    # Check results are equal: Python to Numpy
    for (
        a,
        b,
    ) in zip(index_a_np, index_a):
        assert a == b
    for (
        a,
        b,
    ) in zip(index_b_np, index_b):
        assert a == b


def test_intersection_i_seqs_speed():

    exec_times = []
    for i in range(100):

        size = 10000

        sel1 = flex.random_selection(size, int(size / 2))
        sel2 = flex.random_selection(size, int(size / 2))

        # C++ version
        start = time.perf_counter_ns()
        index_a_cpp, index_b_cpp = intersection_i_seqs_unsorted(sel1, sel2)
        end = time.perf_counter_ns()
        wc_time_cpp = end - start

        # Python version
        start = time.perf_counter_ns()
        index_a = flex.size_t(0)
        index_b = flex.size_t(0)
        lookup = {}
        for i_a, val in enumerate(sel1):
            lookup[val] = i_a
        for i_b, val in enumerate(sel2):
            i_a = lookup.get(val)
            if i_a is not None:
                index_a.append(i_a)
                index_b.append(i_b)
        end = time.perf_counter_ns()
        wc_time_py = end - start

        for (
            a,
            b,
        ) in zip(index_a_cpp, index_a):
            assert a == b
        for (
            a,
            b,
        ) in zip(index_b_cpp, index_b):
            assert a == b

        exec_times.append((wc_time_cpp, wc_time_py))

    tot_cpp, tot_py = zip(*exec_times)
    print(f"Total time in C++ function: {sum(tot_cpp)}")
    print(f"Total time in Python version: {sum(tot_py)}")

    # We use the C++ version because it is faster than the Python version.
    # Let's ensure that's the case across platforms
    assert sum(tot_cpp) / sum(tot_py) < 1.0


@pytest.mark.parametrize("element_type", ["vec3", "mat3"])
def test_ReconstituteDerivatives(element_type):

    if element_type == "vec3":
        n = 3
        build = build_reconstitute_derivatives_vec3
        flex_type = flex.vec3_double
    else:
        n = 9
        build = build_reconstitute_derivatives_mat3
        flex_type = flex.mat3_double

    # Create 100 random (element, indices) pairs, with up to 100 elements in
    # the indices
    pair_data = []
    total_size = 0
    for i in range(100):
        element = [random.random() for j in range(n)]
        indices = flex.random_size_t(random.randint(0, 100))
        pair_data.append((element, indices))
        total_size += indices.size()

    rec_der = build(total_size)

    reference_data = flex_type(0)
    reference_indices = flex.size_t(0)
    for element, indices in pair_data:
        rec_der.add_data(element, indices)
        n = indices.size()
        values = flex_type(n, element)
        reference_data.extend(values)
        reference_indices.extend(indices)

    # Compare the reference values from extending flex arrays with the
    # reconstituted values
    test_data = rec_der.get_data()
    test_indices = rec_der.get_indices()

    assert test_data.as_double().all_eq(reference_data.as_double())
    assert test_indices.all_eq(reference_indices)


if __name__ == "__main__":
    cmdline_overrides = sys.argv[1:]
    test(cmdline_overrides)



from __future__ import annotations

from pathlib import Path

import setuptools

from build import build

__version_tag__ = "3.9.dev"

setup_kwargs = {
    "name": "dials",
    "version": __version_tag__,
    "long_description": Path(__file__).parent.joinpath("README.md").read_text(),
    "description": "Diffraction Integration for Advanced Light Sources",
    "author": "Diamond Light Source",
    "license": "BSD-3-Clause",
    "author_email": "dials-support@lists.sourceforge.net",
    "project_urls": {
        "homepage": "https://dials.github.io",
        "repository": "https://github.com/dials/dials",
    },
    "packages": setuptools.find_packages(where="src"),
    "package_dir": {"": "src"},
    "package_data": {
        "": ["*", "boost_python/*"],
        "dials": [
            "static/katex/*",
            "static/katex/fonts/*",
            "static/katex/contrib/*",
            "static/js/*",
            "static/css/*",
            "templates/*",
        ],
        "dials.algorithms": ["spatial_indexing/*"],
    },
    "classifiers": [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    "entry_points": {
        "libtbx.precommit": ["dials=dials"],
        "libtbx.dispatcher.script": ["pytest=pytest"],
        "dxtbx.profile_model": [
            "gaussian_rs = dials.extensions.gaussian_rs_profile_model_ext:GaussianRSProfileModelExt",
            "ellipsoid = dials.extensions.ellipsoid_profile_model_ext:EllipsoidProfileModelExt",
        ],
        "dxtbx.scaling_model_ext": [
            "physical = dials.algorithms.scaling.model.model:PhysicalScalingModel",
            "KB = dials.algorithms.scaling.model.model:KBScalingModel",
            "array = dials.algorithms.scaling.model.model:ArrayScalingModel",
            "dose_decay = dials.algorithms.scaling.model.model:DoseDecay",
        ],
        "dials.index.basis_vector_search": [
            "fft1d = dials.algorithms.indexing.basis_vector_search:FFT1D",
            "fft3d = dials.algorithms.indexing.basis_vector_search:FFT3D",
            "real_space_grid_search = dials.algorithms.indexing.basis_vector_search:RealSpaceGridSearch",
        ],
        "dials.index.lattice_search": [
            "low_res_spot_match = dials.algorithms.indexing.lattice_search:LowResSpotMatch"
        ],
        "dials.integration.background": [
            "Auto = dials.extensions.auto_background_ext:AutoBackgroundExt",
            "glm = dials.extensions.glm_background_ext:GLMBackgroundExt",
            "gmodel = dials.extensions.gmodel_background_ext:GModelBackgroundExt",
            "simple = dials.extensions.simple_background_ext:SimpleBackgroundExt",
            "null = dials.extensions.null_background_ext:NullBackgroundExt",
        ],
        "dials.integration.centroid": [
            "simple = dials.extensions.simple_centroid_ext:SimpleCentroidExt",
        ],
        "dials.spotfinder.threshold": [
            "dispersion = dials.extensions.dispersion_spotfinder_threshold_ext:DispersionSpotFinderThresholdExt",
            "dispersion_extended = dials.extensions.dispersion_extended_spotfinder_threshold_ext:DispersionExtendedSpotFinderThresholdExt",
            "radial_profile = dials.extensions.radial_profile_spotfinder_threshold_ext:RadialProfileSpotFinderThresholdExt",
        ],
    },
}

build(setup_kwargs)
setuptools.setup(**setup_kwargs)



from __future__ import annotations

from scitbx import matrix

from dials.algorithms.refinement.parameterisation.model_parameters import (
    ModelParameterisation,
    Parameter,
)
from dials.algorithms.refinement.refinement_helpers import dR_from_axis_and_angle


class GoniometerMixin:
    """Mix-in class defining some functionality unique to goniometer
    parameterisations that can be shared by static and scan-varying versions"""

    @staticmethod
    def _build_p_list(axis, beam, parameter_type=Parameter):
        """Build the list of parameters, using the parameter_type callback to
        select between versions of the Parameter class"""

        # Set up the parameters
        if beam:
            s0u = matrix.col(beam.get_unit_s0())
            # define axis orthogonal to spindle-beam plane
            dir2 = axis.cross(s0u).normalize()
            # define axis in the spindle-beam plane
            dir1 = dir2.cross(axis).normalize()
        else:
            dir1 = axis.ortho().normalize()
            dir2 = axis.cross(dir1).normalize()

        # rotation around dir1
        gamma1 = parameter_type(0.0, dir1, "angle (mrad)", "Gamma1")
        # rotation around dir2
        gamma2 = parameter_type(0.0, dir2, "angle (mrad)", "Gamma2")

        # build the parameter list in a specific,  maintained order
        p_list = [gamma1, gamma2]

        return p_list

    @staticmethod
    def _compose_core(iS, gamma1, gamma2, gamma1_axis, gamma2_axis):

        # convert angles to radians
        g1rad, g2rad = gamma1 / 1000.0, gamma2 / 1000.0

        # compose rotation matrices and their first order derivatives
        G1 = (gamma1_axis).axis_and_angle_as_r3_rotation_matrix(g1rad, deg=False)
        dG1_dg1 = dR_from_axis_and_angle(gamma1_axis, g1rad, deg=False)

        G2 = (gamma2_axis).axis_and_angle_as_r3_rotation_matrix(g2rad, deg=False)
        dG2_dg2 = dR_from_axis_and_angle(gamma2_axis, g2rad, deg=False)

        # compose new state
        G21 = G2 * G1
        S = G21 * iS

        # calculate derivatives of S wrt angles:
        #  1) derivative wrt gamma1
        dG21_dg1 = G2 * dG1_dg1
        dS_dg1 = dG21_dg1 * iS

        #  2) derivative wrt gamma2
        dG21_dg2 = dG2_dg2 * G1
        dS_dg2 = dG21_dg2 * iS

        # Convert derivatives back to mrad
        dS_dval = [dS_dg1 / 1000.0, dS_dg2 / 1000.0]

        return S, dS_dval


class GoniometerParameterisation(ModelParameterisation, GoniometerMixin):
    """A parameterisation of a Goniometer model's setting rotation.

    The setting rotation matrix is parameterised using two orientation angles
    expressed in mrad. A beam can be provided (if present in the experiment) to
    ensure a consistent definition of the orientation angles with respect to the
    initial spindle-beam plane in the laboratory frame"""

    def __init__(self, goniometer, beam=None, experiment_ids=None):
        """Initialise the GoniometerParameterisation object

        Args:
            goniometer: A dxtbx Beam object to be parameterised.
            beam: An optional dxtbx Beam object. Defaults to None.
            experiment_ids (list): The experiment IDs affected by this
                parameterisation. Defaults to None, which is replaced by [0].
        """
        # The state of the goniometer model consists of the setting matrix [S] that
        # determines the orientation of the rotation axis in the laboratory frame
        # e_lab = [S] e_datum.
        # This parameters are orientation angles around two axes orthogonal to the
        # initial direction of e_lab. If a beam is supplied, one of these axes will
        # be within the plane containing s0 and e_lab.

        # Set up the initial state
        if experiment_ids is None:
            experiment_ids = [0]
        e_lab = matrix.col(goniometer.get_rotation_axis())
        istate = matrix.sqr(goniometer.get_setting_rotation())

        # build the parameter list
        p_list = self._build_p_list(e_lab, beam)

        # set up the base class
        ModelParameterisation.__init__(
            self, goniometer, istate, p_list, experiment_ids=experiment_ids
        )

        # call compose to calculate all the derivatives
        self.compose()

        return

    def compose(self):

        # extract setting matrix from the initial state
        iS = self._initial_state

        # extract parameters from the internal list
        gamma1, gamma2 = self._param

        # calculate new [S] and derivatives
        S, self._dstate_dp = self._compose_core(
            iS,
            gamma1.value,
            gamma2.value,
            gamma1_axis=gamma1.axis,
            gamma2_axis=gamma2.axis,
        )

        # now update the model with its new [S]
        self._model.set_setting_rotation(S)

        return

    def get_state(self):

        # only a single setting matrix exists, so no multi_state_elt argument is
        # allowed

        return matrix.sqr(self._model.get_setting_rotation())



from __future__ import annotations

from dials_algorithms_image_centroid_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "CentroidImage2dDouble",
    "CentroidImage2dFloat",
    "CentroidImage3dDouble",
    "CentroidImage3dFloat",
    "CentroidMaskedImage2dDouble",
    "CentroidMaskedImage2dFloat",
    "CentroidMaskedImage3dDouble",
    "CentroidMaskedImage3dFloat",
    "CentroidPoints2dDouble",
    "CentroidPoints2dFloat",
    "CentroidPoints3dDouble",
    "CentroidPoints3dFloat",
    "centroid_bias_sq",
    "centroid_image",
    "centroid_points",
)



from __future__ import annotations

import logging
import math

from libtbx.math_utils import iceil

from dials.array_family import flex

logger = logging.getLogger(__name__)


def blank_counts_analysis(reflections, scan, phi_step, fractional_loss):
    if not len(reflections):
        raise ValueError("Input contains no reflections")

    xyz_px = reflections["xyzobs.px.value"]
    x_px, y_px, z_px = xyz_px.parts()
    phi = scan.get_angle_from_array_index(z_px)

    osc = scan.get_oscillation()[1]
    n_images_per_step = iceil(phi_step / osc)
    phi_step = n_images_per_step * osc

    array_range = scan.get_array_range()
    phi_min = scan.get_angle_from_array_index(array_range[0])
    phi_max = scan.get_angle_from_array_index(array_range[1])
    assert phi_min <= flex.min(phi)
    assert phi_max >= flex.max(phi)
    n_steps = max(int(round((phi_max - phi_min) / phi_step)), 1)
    hist = flex.histogram(
        z_px, data_min=array_range[0], data_max=array_range[1], n_slots=n_steps
    )
    logger.debug("Histogram:")
    logger.debug(hist.as_str())

    counts = hist.slots()
    fractional_counts = counts.as_double() / flex.max(counts)

    potential_blank_sel = fractional_counts <= fractional_loss

    xmin, xmax = zip(
        *[
            (slot_info.low_cutoff, slot_info.high_cutoff)
            for slot_info in hist.slot_infos()
        ]
    )

    d = {
        "data": [
            {
                "x": list(hist.slot_centers()),
                "y": list(hist.slots()),
                "xlow": xmin,
                "xhigh": xmax,
                "blank": list(potential_blank_sel),
                "type": "bar",
                "name": "blank_counts_analysis",
            }
        ],
        "layout": {
            "xaxis": {"title": "z observed (images)"},
            "yaxis": {"title": "Number of reflections"},
            "bargap": 0,
        },
    }

    blank_regions = blank_regions_from_sel(d["data"][0])
    d["blank_regions"] = blank_regions

    return d


def blank_integrated_analysis(reflections, scan, phi_step, fractional_loss):
    prf_sel = reflections.get_flags(reflections.flags.integrated_prf)
    if prf_sel.count(True) > 0:
        reflections = reflections.select(prf_sel)
        intensities = reflections["intensity.prf.value"]
        variances = reflections["intensity.prf.variance"]
    else:
        sum_sel = reflections.get_flags(reflections.flags.integrated_sum)
        reflections = reflections.select(sum_sel)
        intensities = reflections["intensity.sum.value"]
        variances = reflections["intensity.sum.variance"]

    i_sigi = intensities / flex.sqrt(variances)

    xyz_px = reflections["xyzobs.px.value"]
    x_px, y_px, z_px = xyz_px.parts()
    phi = scan.get_angle_from_array_index(z_px)

    osc = scan.get_oscillation()[1]
    n_images_per_step = iceil(phi_step / osc)
    phi_step = n_images_per_step * osc

    array_range = scan.get_array_range()
    phi_min = flex.min(phi)
    phi_max = flex.max(phi)
    n_steps = int(round((phi_max - phi_min) / phi_step))
    hist = flex.histogram(
        z_px, data_min=array_range[0], data_max=array_range[1], n_slots=n_steps
    )
    logger.debug("Histogram:")
    logger.debug(hist.as_str())

    mean_i_sigi = flex.double()
    for i, slot_info in enumerate(hist.slot_infos()):
        sel = (z_px >= slot_info.low_cutoff) & (z_px < slot_info.high_cutoff)
        if sel.count(True) == 0:
            mean_i_sigi.append(0)
        else:
            mean_i_sigi.append(flex.mean(i_sigi.select(sel)))

    potential_blank_sel = mean_i_sigi <= (fractional_loss * flex.max(mean_i_sigi))

    xmin, xmax = zip(
        *[
            (slot_info.low_cutoff, slot_info.high_cutoff)
            for slot_info in hist.slot_infos()
        ]
    )

    d = {
        "data": [
            {
                "x": list(hist.slot_centers()),
                "y": list(mean_i_sigi),
                "xlow": xmin,
                "xhigh": xmax,
                "blank": list(potential_blank_sel),
                "type": "bar",
                "name": "blank_counts_analysis",
            }
        ],
        "layout": {
            "xaxis": {"title": "z observed (images)"},
            "yaxis": {"title": "Number of reflections"},
            "bargap": 0,
        },
    }

    blank_regions = blank_regions_from_sel(d["data"][0])
    d["blank_regions"] = blank_regions

    return d


def blank_regions_from_sel(d):
    blank_sel = d["blank"]
    xlow = d["xlow"]
    xhigh = d["xhigh"]

    blank_regions = []
    n = len(blank_sel)

    for i in range(len(blank_sel)):
        if blank_sel[i]:
            if i == 0 or not blank_sel[i - 1]:
                blank_start = math.floor(xlow[i])
            blank_end = math.ceil(xhigh[i])
        if (not blank_sel[i] and i > 0 and blank_sel[i - 1]) or (
            i == (n - 1) and blank_sel[i - 1]
        ):
            blank_regions.append((blank_start, blank_end))

    return blank_regions



from __future__ import annotations

import procrunner


def test(dials_data, tmp_path):
    images = dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")

    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            "output.experiments=spotfinder.expt",
            "output.reflections=spotfinder.refl",
            "output.shoeboxes=True",
        ]
        + sorted(images),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("spotfinder.expt").is_file()
    assert tmp_path.joinpath("spotfinder.refl").is_file()

    result = procrunner.run(
        [
            "dials.find_hot_pixels",
            "input.experiments=spotfinder.expt",
            "input.reflections=spotfinder.refl",
            "output.mask=hot_pixels.mask",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("hot_pixels.mask").is_file()
    assert (
        b"Found 8 hot pixels" in result.stdout or b"Found 9 hot pixels" in result.stdout
    )



from __future__ import annotations

from dials.algorithms.background.glm.algorithm import BackgroundAlgorithm
from dials_algorithms_background_glm_ext import *  # noqa: F403; lgtm

__all__ = ("BackgroundAlgorithm", "Creator", "RobustPoissonMean")  # noqa: F405



# DIALS_ENABLE_COMMAND_LINE_COMPLETION


from __future__ import annotations

import copy
import os
import sys

import iotbx.phil
from cctbx import sgtbx
from rstbx.symmetry.constraints import parameter_reduction

import dials.util
from dials.algorithms.indexing.assign_indices import AssignIndicesGlobal
from dials.algorithms.scaling.scaling_library import determine_best_unit_cell
from dials.array_family import flex
from dials.util.filter_reflections import filtered_arrays_from_experiments_reflections
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files

help_message = """

This program can be used to re-index an indexed.expt and/or indexed.refl
file from one setting to another. The change of basis operator can be
provided in h,k,l, or a,b,c or x,y,z conventions. By default the change of
basis operator will also be applied to the space group in the indexed.expt
file, however, optionally, a space group (including setting) to be applied
AFTER applying the change of basis operator can be provided.
Alternatively, to reindex an integated dataset in the case of indexing ambiguity,
a reference dataset (models.expt and reflection.refl) in the same space
group can be specified. In this case, any potential twin operators are tested,
and the dataset is reindexed to the setting that gives the highest correlation
with the reference dataset.

Examples::

  dials.reindex indexed.expt change_of_basis_op=b+c,a+c,a+b

  dials.reindex indexed.refl change_of_basis_op=-b,a+b+2*c,-a

  dials.reindex indexed.expt indexed.refl change_of_basis_op=l,h,k

  dials.reindex indexed.expt indexed.refl reference.experiments=reference.expt
    reference.reflections=reference.refl
"""

phil_scope = iotbx.phil.parse(
    """
change_of_basis_op = a,b,c
  .type = str
hkl_offset = None
  .type = ints(size=3)
space_group = None
  .type = space_group
  .help = "The space group to be applied AFTER applying the change of basis "
           "operator."
reference {
  experiments = None
    .type = path
    .help = "Reference experiment for determination of change of basis operator."
  reflections = None
    .type = path
    .help = "Reference reflections to allow reindexing to consistent index between datasets."
}
output {
  experiments = reindexed.expt
    .type = str
    .help = "The filename for reindexed experimental models"

  reflections = reindexed.refl
    .type = str
    .help = "The filename for reindexed reflections"
}
""",
    process_includes=True,
)


def derive_change_of_basis_op(from_hkl, to_hkl):

    # exclude those reflections that we couldn't index
    sel = (to_hkl != (0, 0, 0)) & (from_hkl != (0, 0, 0))
    assert sel.count(True) >= 3  # need minimum of 3 equations ?
    to_hkl = to_hkl.select(sel)
    from_hkl = from_hkl.select(sel)

    # for each miller index, solve a system of linear equations to find the
    # change of basis operator
    h, k, l = to_hkl.as_vec3_double().parts()

    r = []
    from scitbx.lstbx import normal_eqns

    for i in range(3):
        eqns = normal_eqns.linear_ls(3)
        for index, hkl in zip((h, k, l)[i], from_hkl):
            eqns.add_equation(
                right_hand_side=index, design_matrix_row=flex.double(hkl), weight=1
            )
        eqns.solve()
        r.extend(eqns.solution())

    from scitbx import matrix
    from scitbx.math import continued_fraction

    denom = 12
    r = [
        int(denom * continued_fraction.from_real(r_, eps=1e-2).as_rational())
        for r_ in r
    ]
    r = matrix.sqr(r).transpose()
    # print (1/denom)*r

    # now convert into a cctbx change_of_basis_op object
    change_of_basis_op = sgtbx.change_of_basis_op(
        sgtbx.rt_mx(sgtbx.rot_mx(r, denominator=denom))
    ).inverse()
    print(f"discovered change_of_basis_op={change_of_basis_op}")

    # sanity check that this is the right cb_op
    assert (change_of_basis_op.apply(from_hkl) == to_hkl).count(False) == 0

    return change_of_basis_op


def reindex_experiments(experiments, cb_op, space_group=None):
    reindexed_experiments = copy.deepcopy(experiments)

    for crystal in reindexed_experiments.crystals():
        cryst_reindexed = copy.deepcopy(crystal)
        if space_group is not None:
            # See also https://github.com/cctbx/cctbx_project/issues/424
            cryst_reindexed.set_space_group(sgtbx.space_group("P 1"))
            cryst_reindexed = cryst_reindexed.change_basis(cb_op)
            cryst_reindexed.set_space_group(space_group)
            S = parameter_reduction.symmetrize_reduce_enlarge(
                cryst_reindexed.get_space_group()
            )
            S.set_orientation(cryst_reindexed.get_B())
            S.symmetrize()
            # Cache the scan-varying A matrices if applicable as these get lost
            # when we call crystal.set_B()
            A_varying = [
                cryst_reindexed.get_A_at_scan_point(i)
                for i in range(cryst_reindexed.num_scan_points)
            ]
            # Update the symmetrized B matrix
            cryst_reindexed.set_B(S.orientation.reciprocal_matrix())
            # Reapply the scan-varying A matrices
            cryst_reindexed.set_A_at_scan_points(A_varying)
        else:
            cryst_reindexed = cryst_reindexed.change_basis(cb_op)
        crystal.update(cryst_reindexed)

    return reindexed_experiments


@dials.util.show_mail_handle_errors()
def run(args=None):
    import libtbx.load_env

    from dials.util import Sorry

    usage = "dials.reindex [options] indexed.expt indexed.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )
    if len(experiments) == 0 and len(reflections) == 0:
        parser.print_help()
        return
    if params.change_of_basis_op is None:
        raise Sorry("Please provide a change_of_basis_op.")

    reference_crystal = None
    if params.reference.experiments is not None:
        from dxtbx.serialize import load

        reference_experiments = load.experiment_list(
            params.reference.experiments, check_format=False
        )
        if len(reference_experiments.crystals()) == 1:
            reference_crystal = reference_experiments.crystals()[0]
        else:
            # first check sg all same
            sgs = [
                expt.crystal.get_space_group().type().number() for expt in experiments
            ]
            if len(set(sgs)) > 1:
                raise Sorry(
                    """The reference experiments have different space groups:
                    space group numbers found: %s
                    Please reanalyse the data so that space groups are consistent,
                    (consider using dials.reindex, dials.symmetry or dials.cosym)"""
                    % ", ".join(map(str, set(sgs)))
                )

            reference_crystal = reference_experiments.crystals()[0]
            reference_crystal.unit_cell = determine_best_unit_cell(
                reference_experiments
            )

    if params.reference.reflections is not None:
        # First check that we have everything as expected for the reference reindexing
        if params.reference.experiments is None:
            raise Sorry(
                """For reindexing against a reference dataset, a reference
experiments file must also be specified with the option: reference.experiments= """
            )
        if not os.path.exists(params.reference.reflections):
            raise Sorry("Could not locate reference dataset reflection file")

        reference_reflections = flex.reflection_table().from_file(
            params.reference.reflections
        )

        test_reflections = reflections[0]

        if (
            reference_crystal.get_space_group().type().number()
            != experiments.crystals()[0].get_space_group().type().number()
        ):
            raise Sorry("Space group of input does not match reference")

        # Set some flags to allow filtering, if wanting to reindex against
        # reference with data that has not yet been through integration
        if (
            test_reflections.get_flags(test_reflections.flags.integrated_sum).count(
                True
            )
            == 0
        ):
            assert (
                "intensity.sum.value" in test_reflections
            ), "No 'intensity.sum.value' in reflections"
            test_reflections.set_flags(
                flex.bool(test_reflections.size(), True),
                test_reflections.flags.integrated_sum,
            )
        if (
            reference_reflections.get_flags(
                reference_reflections.flags.integrated_sum
            ).count(True)
            == 0
        ):
            assert (
                "intensity.sum.value" in test_reflections
            ), "No 'intensity.sum.value in reference reflections"
            reference_reflections.set_flags(
                flex.bool(reference_reflections.size(), True),
                reference_reflections.flags.integrated_sum,
            )

        # Make miller array of the two datasets
        try:
            test_miller_set = filtered_arrays_from_experiments_reflections(
                experiments, [test_reflections]
            )[0]
        except ValueError:
            raise Sorry("No reflections remain after filtering the test dataset")
        try:
            reference_miller_set = filtered_arrays_from_experiments_reflections(
                reference_experiments, [reference_reflections]
            )[0]
        except ValueError:
            raise Sorry("No reflections remain after filtering the reference dataset")

        from dials.algorithms.symmetry.reindex_to_reference import (
            determine_reindex_operator_against_reference,
        )

        change_of_basis_op = determine_reindex_operator_against_reference(
            test_miller_set, reference_miller_set
        )

    elif len(experiments) and params.change_of_basis_op is libtbx.Auto:
        if reference_crystal is not None:
            if len(experiments.crystals()) > 1:
                raise Sorry("Only one crystal can be processed at a time")
            from dials.algorithms.indexing.compare_orientation_matrices import (
                difference_rotation_matrix_axis_angle,
            )

            cryst = experiments.crystals()[0]
            R, axis, angle, change_of_basis_op = difference_rotation_matrix_axis_angle(
                cryst, reference_crystal
            )
            print(f"Change of basis op: {change_of_basis_op}")
            print("Rotation matrix to transform input crystal to reference::")
            print(R.mathematica_form(format="%.3f", one_row_per_line=True))
            print(
                f"Rotation of {angle:.3f} degrees",
                "about axis (%.3f, %.3f, %.3f)" % axis,
            )

        elif len(reflections):
            assert len(reflections) == 1

            # always re-map reflections to reciprocal space
            refl = reflections.deep_copy()
            refl.centroid_px_to_mm(experiments)
            refl.map_centroids_to_reciprocal_space(experiments)

            # index the reflection list using the input experiments list
            refl["id"] = flex.int(len(refl), -1)
            index = AssignIndicesGlobal(tolerance=0.2)
            index(refl, experiments)
            hkl_expt = refl["miller_index"]
            hkl_input = reflections[0]["miller_index"]

            change_of_basis_op = derive_change_of_basis_op(hkl_input, hkl_expt)

            # reset experiments list since we don't want to reindex this
            experiments = []

    else:
        change_of_basis_op = sgtbx.change_of_basis_op(params.change_of_basis_op)

    if len(experiments):
        space_group = params.space_group
        if space_group is not None:
            space_group = space_group.group()
        try:
            experiments = reindex_experiments(
                experiments, change_of_basis_op, space_group=space_group
            )
        except RuntimeError as e:
            # Only catch specific errors here
            if "Unsuitable value for rational rotation matrix." in str(e):
                original_message = str(e).split(":")[-1].strip()
                sys.exit(f"Error: {original_message} Is your change_of_basis_op valid?")
            elif "DXTBX_ASSERT(detail::is_r3_rotation_matrix(U)) failure" in str(e):
                sys.exit(
                    "Error: U is not a rotation matrix. Is your change_of_basis_op valid?"
                )
            raise

        print(f"Saving reindexed experimental models to {params.output.experiments}")
        experiments.as_file(params.output.experiments)

    if len(reflections):
        assert len(reflections) == 1
        reflections = reflections[0]

        miller_indices = reflections["miller_index"]

        if params.hkl_offset is not None:
            h, k, l = miller_indices.as_vec3_double().parts()
            h += params.hkl_offset[0]
            k += params.hkl_offset[1]
            l += params.hkl_offset[2]
            miller_indices = flex.miller_index(h.iround(), k.iround(), l.iround())
        non_integral_indices = change_of_basis_op.apply_results_in_non_integral_indices(
            miller_indices
        )
        if non_integral_indices.size() > 0:
            print(
                "Removing %i/%i reflections (change of basis results in non-integral indices)"
                % (non_integral_indices.size(), miller_indices.size())
            )
        sel = flex.bool(miller_indices.size(), True)
        sel.set_selected(non_integral_indices, False)
        miller_indices_reindexed = change_of_basis_op.apply(miller_indices.select(sel))
        reflections["miller_index"].set_selected(sel, miller_indices_reindexed)
        reflections["miller_index"].set_selected(~sel, (0, 0, 0))

        print(f"Saving reindexed reflections to {params.output.reflections}")
        reflections.as_file(params.output.reflections)


if __name__ == "__main__":
    run()



from __future__ import annotations

import logging

logging.getLogger("dials").addHandler(logging.NullHandler())

# Intercept easy_mp exceptions to extract stack traces before they are lost at
# the libtbx process boundary/the easy_mp API. In the case of a subprocess
# crash we print the subprocess stack trace, which will be most useful for
# debugging parallelized sections of DIALS code.
import libtbx.scheduling.stacktrace as _lss


def _stacktrace_tracer(error, trace, intercepted_call=_lss.set_last_exception):
    """Intercepts and prints ephemeral stacktraces."""
    if error and trace:
        logging.getLogger("dials").error(
            "\n\neasy_mp crash detected; subprocess trace: ----\n%s%s\n%s\n\n",
            "".join(trace),
            error,
            "-" * 46,
        )
    return intercepted_call(error, trace)


if _lss.set_last_exception.__doc__ != _stacktrace_tracer.__doc__:
    # ensure function is only redirected once
    _lss.set_last_exception = _stacktrace_tracer



"""
Refinement engine and functions for error model refinement.
"""
from __future__ import annotations

import logging

from dials.algorithms.refinement.engine import SimpleLBFGS
from dials.algorithms.scaling.error_model.error_model import (
    ErrorModelA_APM,
    ErrorModelB_APM,
    ErrorModelRegressionAPM,
)
from dials.algorithms.scaling.error_model.error_model_target import (
    ErrorModelTargetA,
    ErrorModelTargetB,
    ErrorModelTargetRegression,
)
from dials.algorithms.scaling.scaling_refiner import print_step_table

logger = logging.getLogger("dials")


def run_error_model_refinement(model, Ih_table):
    """
    Refine an error model for the input data, returning the model.

    Raises:
        ValueError: if insufficient reflections or bad refined value.
        RuntimeError: can be raised in LBFGS minimiser.
    """
    assert Ih_table.n_work_blocks == 1
    model.configure_for_refinement(Ih_table.blocked_data_list[0])
    if not model.active_parameters:
        logger.info("All error model parameters fixed, skipping refinement")
    else:
        logger.info("Performing a round of error model refinement.")
        refinery = error_model_refinery(
            model=model,
            active_parameters=model.active_parameters,
            error_model_scope=model.params,
            max_iterations=100,
        )
        if refinery:
            refinery.run()
            refinery.print_step_table()
        logger.info(model)
    model.finalise()
    return model


def error_model_refinery(model, active_parameters, error_model_scope, max_iterations):
    """Return the correct engine based on phil parameters.

    Note that here the target also takes the role of the predication
    parameterisation by implementing the set_param_vals and get_param_vals
    methods (the code is organised in this way to allow the use of the
    dials.refinement engines)."""
    if model.id_ == "basic":
        if error_model_scope.minimisation == "individual":
            return ErrorModelRefinery(
                model=model,
                max_iterations=max_iterations,
                parameters_to_refine=active_parameters,
            )
        if error_model_scope.minimisation == "regression":
            parameterisation = ErrorModelRegressionAPM(model, active_parameters)
            target = ErrorModelTargetRegression(model)
            return ErrorModelRegressionRefiner(
                model=model,
                target=target,
                prediction_parameterisation=parameterisation,
                max_iterations=max_iterations,
            )
        else:
            return None


class ErrorModelRegressionRefiner(SimpleLBFGS):

    """Use LBFGS for convenience, actually is a linear regression.

    Therefore target.predict step is unnecessary."""

    def __init__(self, model, *args, **kwargs):
        self.model = model
        self.parameterisation = kwargs["prediction_parameterisation"]
        SimpleLBFGS.__init__(self, *args, **kwargs)

    def print_step_table(self):
        print_step_table(self)

    def update_journal(self):
        """Append latest step information to the journal attributes"""

        # add step quantities to journal
        self.history.add_row()
        self.history.set_last_cell("num_reflections", self.model.n_refl)
        self.history.set_last_cell("rmsd", self._target.rmsds(self.parameterisation))
        self.history.set_last_cell(
            "parameter_vector", self.parameterisation.get_param_vals()
        )
        self.history.set_last_cell("objective", self._f)
        if "gradient" in self.history:
            self.history.set_last_cell("gradient", self._g)
        return

    def run(self):
        super().run()
        self.parameterisation.resolve_model_parameters()

    def prepare_for_step(self):
        """Update the parameterisation and prepare the target function. Overwrites
        the prepare_for_step method from refinery to direct the updating away from
        the target function to the update_for_minimisation method."""

        x = self.x

        # set current parameter values
        self.parameterisation.set_param_vals(x)

        return

    def compute_functional_gradients_and_curvatures(self):
        """overwrite method to avoid calls to 'blocks' methods of target"""
        logger.debug("Current parameters %s", [f"{i:.6f}" for i in self.x])
        self.prepare_for_step()
        self._target.predict(
            self.parameterisation
        )  # null operation as linear regression

        f, g = self._target.compute_functional_gradients(self.parameterisation)

        # restraints terms
        restraints = self._target.compute_restraints_functional_gradients(
            self.parameterisation
        )

        if restraints:
            f += restraints[0]
            g += restraints[1]
        logger.debug("Current functional %s", f)
        return f, g, None


class ErrorModelRefinery:

    """Refiner for the basic error model."""

    def __init__(self, model, parameters_to_refine, *args, **kwargs):
        self.model = model
        self.parameters_to_refine = parameters_to_refine
        self.args = args
        self.kwargs = kwargs
        self.avals = []
        self.bvals = []
        self._avals_tolerance = 0.01
        self.converged = False

    def print_step_table(self):
        pass

    def test_value_convergence(self):
        """Test for convergence of RMSDs"""

        # http://en.wikipedia.org/wiki/
        # Non-linear_least_squares#Convergence_criteria
        try:
            r1 = self.avals[-1]
            r2 = self.avals[-2]
        except IndexError:
            return False

        if r2 > 0:
            return abs((r2 - r1) / r2) < self._avals_tolerance
        else:
            return True

    def _refine_a(self):
        parameterisation = ErrorModelA_APM(self.model)
        self._refine_component(
            self.model, ErrorModelTargetA(self.model), parameterisation
        )

    def _refine_b(self):
        parameterisation = ErrorModelB_APM(self.model)
        self._refine_component(
            self.model, ErrorModelTargetB(self.model), parameterisation
        )

    def _refine_component(self, model, target, parameterisation):
        refiner = ErrorModelRegressionRefiner(
            model=model,
            target=target,
            prediction_parameterisation=parameterisation,
            *self.args,
            **self.kwargs,
        )
        refiner.run()

    def run(self):
        """Refine the model."""
        if self.parameters_to_refine == ["a", "b"]:
            for n in range(20):  # usually converges in around 5 cycles
                self._refine_a()
                # now update in model
                self.avals.append(self.model.components["a"].parameters[0])
                logger.debug(
                    "Error model refinement cycle %s: a = %s", n, self.avals[-1]
                )
                self.model.update(self.model.parameters)
                self._refine_b()
                self.bvals.append(self.model.components["b"].parameters[0])
                logger.debug(
                    "Error model refinement cycle %s: b = %s", n, self.bvals[-1]
                )
                self.model.update(self.model.parameters)
                if self.test_value_convergence():
                    self.converged = True
                    break
            if self.converged and self.avals[-1] > 0.4:
                self._refine_a()
                self.avals.append(self.model.components["a"].parameters[0])
                self.model.update(self.model.parameters)
            else:
                logger.info(
                    """
Two-parameter Error model refinement failed.
Performing error model refinement with fixed a=1.0
"""
                )
                self.model.parameters = [1.0, 0.02]
                self._refine_b()
                self.model.update(self.model.parameters)
        elif self.parameters_to_refine == ["a"]:
            self._refine_a()
            self.model.update(self.model.parameters)
        elif self.parameters_to_refine == ["b"]:
            self._refine_b()
            self.model.update(self.model.parameters)



from __future__ import annotations

import random

from dials.algorithms.polygon.spatial_interpolation import (
    regrid_grid_to_irregular_grid,
    regrid_irregular_grid_to_grid,
)


class TestRegridIrregularToRegular:
    @staticmethod
    def test_identical():
        from scitbx.array_family import flex

        # Set the size of the grid
        height = 10
        width = 10

        # Create the grid data
        grid = flex.double([random.uniform(0, 100) for i in range(height * width)])
        grid.reshape(flex.grid(height, width))

        # Create the grid coordinates
        xy = []
        for j in range(height + 1):
            for i in range(width + 1):
                xy.append((i, j))
        gridxy = flex.vec2_double(xy)
        gridxy.reshape(flex.grid(height + 1, width + 1))

        # Get the output grid
        output = regrid_irregular_grid_to_grid(grid, gridxy, (height, width))

        # Check that each pixel is identical
        eps = 1e-7
        for j in range(height):
            for i in range(width):
                assert abs(output[j, i] - grid[j, i]) <= eps

    @staticmethod
    def test_known_offset():
        from scitbx.array_family import flex

        # Set the size of the grid
        height = 10
        width = 10

        # Create the grid data
        grid = flex.double([1 for i in range(height * width)])
        grid.reshape(flex.grid(height, width))

        # Create the grid coordinates
        xy = []
        for j in range(height + 1):
            for i in range(width + 1):
                xy.append((i + 0.5, j + 0.5))
        gridxy = flex.vec2_double(xy)
        gridxy.reshape(flex.grid(height + 1, width + 1))

        # Get the output grid
        output = regrid_irregular_grid_to_grid(grid, gridxy, (height, width))

        # Check that each each pixel along the left and bottom has a value
        # of 0.5 and that everything else is 1
        eps = 1e-7
        assert abs(output[0, 0] - 0.25) <= eps
        for i in range(1, width):
            assert abs(output[0, i] - 0.5) <= eps
        for j in range(1, height):
            assert abs(output[j, 0] - 0.5) <= eps
        for j in range(1, height):
            for i in range(1, width):
                assert abs(output[j, i] - 1.0) <= eps

    @staticmethod
    def test_larger_output():
        from scitbx.array_family import flex

        # Set the size of the grid
        height = 10
        width = 10

        # Create the grid data
        value = 13
        grid = flex.double([value for i in range(height * width)])
        grid.reshape(flex.grid(height, width))

        # Create the grid coordinates
        xy = []
        for j in range(height + 1):
            for i in range(width + 1):
                xy.append((i * 2, j * 2))
        gridxy = flex.vec2_double(xy)
        gridxy.reshape(flex.grid(height + 1, width + 1))

        # Get the output grid
        output = regrid_irregular_grid_to_grid(grid, gridxy, (height * 2, width * 2))

        # Check that each each pixel has a value of 0.25 the input
        eps = 1e-7
        for j in range(1, height):
            for i in range(1, width):
                assert abs(output[j, i] - 0.25 * value) <= eps

    @staticmethod
    def test_larger_input():
        from scitbx.array_family import flex

        # Set the size of the grid
        input_height = 20
        input_width = 20
        output_height = 10
        output_width = 10

        # Create the grid data
        value = 13
        grid = flex.double([value for i in range(input_height * input_width)])
        grid.reshape(flex.grid(input_height, input_width))

        # Create the grid coordinates
        xy = []
        for j in range(input_height + 1):
            for i in range(input_width + 1):
                xy.append((i / 2.0, j / 2.0))
        gridxy = flex.vec2_double(xy)
        gridxy.reshape(flex.grid(input_height + 1, input_width + 1))

        # Get the output grid
        output = regrid_irregular_grid_to_grid(
            grid, gridxy, (output_height, output_width)
        )

        # Check that each each pixel has a value of 4 times the input
        eps = 1e-7
        for j in range(1, output_height):
            for i in range(1, output_width):
                assert abs(output[j, i] - 4 * value) <= eps

    @staticmethod
    def test_known_orientation():
        from math import cos, pi, sin, sqrt

        from scitbx import matrix
        from scitbx.array_family import flex

        # Set the size of the grid
        input_height = 4
        input_width = 4
        output_height = 4
        output_width = 4

        # Create the grid data
        value = 13
        grid = flex.double([value for i in range(input_height * input_width)])
        grid.reshape(flex.grid(input_height, input_width))

        # Create the grid coordinates
        xy = []
        R = matrix.sqr((cos(pi / 4), -sin(pi / 4), sin(pi / 4), cos(pi / 4)))
        for j in range(input_height + 1):
            for i in range(input_width + 1):
                ij = R * matrix.col((i * sqrt(8) / 4, j * sqrt(8) / 4))
                xy.append((ij[0] + 2, ij[1]))
        gridxy = flex.vec2_double(xy)
        gridxy.reshape(flex.grid(input_height + 1, input_width + 1))

        # Get the output grid
        output = regrid_irregular_grid_to_grid(
            grid, gridxy, (output_height, output_width)
        )

        expected = [[0, 1, 1, 0], [1, 2, 2, 1], [1, 2, 2, 1], [0, 1, 1, 0]]

        # Check that each each pixel has a value of the input
        eps = 1e-7
        for j in range(1, output_height):
            for i in range(1, output_width):
                assert abs(output[j, i] - expected[j][i] * value) <= eps

    @staticmethod
    def test_conservation_of_counts():
        from math import cos, pi, sin

        from scitbx import matrix
        from scitbx.array_family import flex

        # Set the size of the grid
        input_height = 10
        input_width = 10
        output_height = 50
        output_width = 50

        # Create the grid data
        grid = flex.double(
            [random.uniform(0, 100) for i in range(input_height * input_width)]
        )
        grid.reshape(flex.grid(input_height, input_width))

        # Create the grid coordinates
        xy = []
        angle = random.uniform(0, pi)
        offset = (random.uniform(20, 30), random.uniform(20, 30))
        R = matrix.sqr((cos(angle), -sin(angle), sin(angle), cos(angle)))
        for j in range(input_height + 1):
            for i in range(input_width + 1):
                ij = R * matrix.col((i, j))
                xy.append((ij[0] + offset[0], ij[1] + offset[0]))
        gridxy = flex.vec2_double(xy)
        gridxy.reshape(flex.grid(input_height + 1, input_width + 1))

        # Get the output grid
        output = regrid_irregular_grid_to_grid(
            grid, gridxy, (output_height, output_width)
        )

        # Check that the sum of the counts is conserved
        eps = 1e-7
        assert abs(flex.sum(output) - flex.sum(grid)) <= eps


class TestRegridRegularToIrregular:
    @staticmethod
    def test_identical():
        from scitbx.array_family import flex

        # Set the size of the grid
        height = 10
        width = 10

        # Create the grid data
        grid = flex.double([random.uniform(0, 100) for i in range(height * width)])
        grid.reshape(flex.grid(height, width))

        # Create the grid coordinates
        xy = []
        for j in range(height + 1):
            for i in range(width + 1):
                xy.append((i, j))
        gridxy = flex.vec2_double(xy)
        gridxy.reshape(flex.grid(height + 1, width + 1))

        # Get the output grid
        output = regrid_grid_to_irregular_grid(grid, gridxy)

        # Check that each pixel is identical
        eps = 1e-7
        for j in range(height):
            for i in range(width):
                assert abs(output[j, i] - grid[j, i]) <= eps

    @staticmethod
    def test_known_offset():
        from scitbx.array_family import flex

        # Set the size of the grid
        height = 10
        width = 10

        # Create the grid data
        grid = flex.double([1 for i in range(height * width)])
        grid.reshape(flex.grid(height, width))

        # Create the grid coordinates
        xy = []
        for j in range(height + 1):
            for i in range(width + 1):
                xy.append((i + 0.5, j + 0.5))
        gridxy = flex.vec2_double(xy)
        gridxy.reshape(flex.grid(height + 1, width + 1))

        # Get the output grid
        output = regrid_grid_to_irregular_grid(grid, gridxy)

        # Check that each each pixel along the left and bottom has a value
        # of 0.5 and that everything else is 1
        eps = 1e-7
        assert abs(output[height - 1, width - 1] - 0.25) <= eps
        for i in range(0, width - 1):
            assert abs(output[height - 1, i] - 0.5) <= eps
        for j in range(0, height - 1):
            assert abs(output[j, width - 1] - 0.5) <= eps
        for j in range(0, height - 1):
            for i in range(0, width - 1):
                assert abs(output[j, i] - 1.0) <= eps

    @staticmethod
    def test_larger_output():
        from scitbx.array_family import flex

        # Set the size of the grid
        height = 10
        width = 10

        # Create the grid data
        value = 13
        grid = flex.double([value for i in range(height * width)])
        grid.reshape(flex.grid(height, width))

        # Create the grid coordinates
        xy = []
        for j in range(2 * height + 1):
            for i in range(2 * width + 1):
                xy.append((i / 2, j / 2))
        gridxy = flex.vec2_double(xy)
        gridxy.reshape(flex.grid(2 * height + 1, 2 * width + 1))

        # Get the output grid
        output = regrid_grid_to_irregular_grid(grid, gridxy)

        # Check that each each pixel has a value of 0.25 the input
        eps = 1e-7
        for j in range(1, height):
            for i in range(1, width):
                assert abs(output[j, i] - 0.25 * value) <= eps

    @staticmethod
    def test_larger_input():
        from scitbx.array_family import flex

        # Set the size of the grid
        input_height = 20
        input_width = 20
        output_height = 10
        output_width = 10

        # Create the grid data
        value = 13
        grid = flex.double([value for i in range(input_height * input_width)])
        grid.reshape(flex.grid(input_height, input_width))

        # Create the grid coordinates
        xy = []
        for j in range(output_height + 1):
            for i in range(output_width + 1):
                xy.append((i * 2, j * 2))
        gridxy = flex.vec2_double(xy)
        gridxy.reshape(flex.grid(output_height + 1, output_width + 1))

        # Get the output grid
        output = regrid_grid_to_irregular_grid(grid, gridxy)

        # Check that each each pixel has a value of 4 times the input
        eps = 1e-7
        for j in range(1, output_height):
            for i in range(1, output_width):
                assert abs(output[j, i] - 4 * value) <= eps

    @staticmethod
    def test_known_orientation():
        from math import cos, pi, sin, sqrt

        from scitbx import matrix
        from scitbx.array_family import flex

        # Set the size of the grid
        input_height = 4
        input_width = 4
        output_height = 4
        output_width = 4

        # Create the grid data
        value = 13
        grid = flex.double([value for i in range(input_height * input_width)])
        grid.reshape(flex.grid(input_height, input_width))

        # Create the grid coordinates
        xy = []
        R = matrix.sqr((cos(pi / 4), -sin(pi / 4), sin(pi / 4), cos(pi / 4)))
        for j in range(output_height + 1):
            for i in range(output_width + 1):
                ij = R * matrix.col((i * sqrt(8) / 2, j * sqrt(8) / 2))
                xy.append((ij[0] + 2, ij[1] - 2))
        gridxy = flex.vec2_double(xy)
        gridxy.reshape(flex.grid(output_height + 1, output_width + 1))

        # Get the output grid
        output = regrid_grid_to_irregular_grid(grid, gridxy)

        expected = [[0, 1, 1, 0], [1, 2, 2, 1], [1, 2, 2, 1], [0, 1, 1, 0]]

        # Check that each each pixel has a value of the input
        eps = 1e-7
        for j in range(1, output_height):
            for i in range(1, output_width):
                assert abs(output[j, i] - expected[j][i] * value) <= eps

    @staticmethod
    def test_conservation_of_counts():
        from math import cos, pi, sin

        from scitbx import matrix
        from scitbx.array_family import flex

        # Set the size of the grid
        input_height = 10
        input_width = 10
        output_height = 50
        output_width = 50

        # Create the grid data
        grid = flex.double(
            [random.uniform(0, 100) for i in range(input_height * input_width)]
        )
        grid.reshape(flex.grid(input_height, input_width))

        # Create the grid coordinates
        xy = []
        angle = random.uniform(0, pi)
        R = matrix.sqr((cos(angle), -sin(angle), sin(angle), cos(angle)))
        for j in range(output_height + 1):
            for i in range(output_width + 1):
                ij = R * matrix.col((i, j))
                xy.append((ij[0], ij[1]))
        gridxy = flex.vec2_double(xy)
        gridxy.reshape(flex.grid(output_height + 1, output_width + 1))
        off = gridxy[output_height // 2, output_width // 2]
        for i in range(len(gridxy)):
            gridxy[i] = (gridxy[i][0] - off[0], gridxy[i][1] - off[1])

        # Get the output grid
        output = regrid_grid_to_irregular_grid(grid, gridxy)
        # Check that the sum of the counts is conserved
        eps = 1e-7
        assert abs(flex.sum(output) - flex.sum(grid)) <= eps



from __future__ import annotations

import os

from cctbx import sgtbx
from dxtbx.model import Crystal
from dxtbx.model.experiment_list import ExperimentListFactory
from iotbx.xds import integrate_hkl, spot_xds
from libtbx.phil import parse
from rstbx.cftbx.coordinate_frame_helpers import align_reference_frame
from scitbx import matrix

from dials.array_family import flex
from dials.util import show_mail_handle_errors
from dials.util.command_line import Command
from dials.util.options import ArgumentParser


class SpotXDSImporter:
    """Class to import a spot.xds file to a reflection table."""

    def __init__(self, spot_xds):
        self._spot_xds = spot_xds

    def __call__(self, params, options):
        """Import the spot.xds file."""
        # Read the SPOT.XDS file
        Command.start("Reading SPOT.XDS")
        handle = spot_xds.reader()
        handle.read_file(self._spot_xds)
        centroid = handle.centroid
        intensity = handle.intensity
        try:
            miller_index = handle.miller_index
        except AttributeError:
            miller_index = None
        Command.end(f"Read {len(centroid)} spots from SPOT.XDS file.")

        # Create the reflection list
        Command.start("Creating reflection list")
        table = flex.reflection_table()
        table["id"] = flex.int(len(centroid), 0)
        table["panel"] = flex.size_t(len(centroid), 0)
        if miller_index:
            table["miller_index"] = flex.miller_index(miller_index)
        table["xyzobs.px.value"] = flex.vec3_double(centroid)
        table["intensity.sum.value"] = flex.double(intensity)
        Command.end("Created reflection list")

        # Remove invalid reflections
        Command.start("Removing invalid reflections")
        if miller_index and params.remove_invalid:
            flags = flex.bool([h != (0, 0, 0) for h in table["miller_index"]])
            table = table.select(flags)
        Command.end(f"Removed invalid reflections, {len(table)} remaining")

        # Fill empty standard columns
        if params.add_standard_columns:
            Command.start("Adding standard columns")
            rt = flex.reflection_table.empty_standard(len(table))
            rt.update(table)
            table = rt
            # set variances to unity
            table["xyzobs.mm.variance"] = flex.vec3_double(len(table), (1, 1, 1))
            table["xyzobs.px.variance"] = flex.vec3_double(len(table), (1, 1, 1))
            Command.end("Standard columns added")

        # Output the table to pickle file
        if params.output.filename is None:
            params.output.filename = "spot_xds.refl"
        Command.start(f"Saving reflection table to {params.output.filename}")
        table.as_file(params.output.filename)
        Command.end(f"Saved reflection table to {params.output.filename}")


class IntegrateHKLImporter:
    """Class to import an integrate.hkl file to a reflection table."""

    def __init__(self, integrate_hkl, experiment):
        self._integrate_hkl = integrate_hkl
        self._experiment = experiment

    def __call__(self, params, options):
        """Import the integrate.hkl file."""
        # Get the unit cell to calculate the resolution
        uc = self._experiment.crystal.get_unit_cell()

        # Read the INTEGRATE.HKL file
        Command.start("Reading INTEGRATE.HKL")
        handle = integrate_hkl.reader()
        handle.read_file(self._integrate_hkl)
        hkl = flex.miller_index(handle.hkl)
        xyzcal = flex.vec3_double(handle.xyzcal)
        xyzobs = flex.vec3_double(handle.xyzobs)
        iobs = flex.double(handle.iobs)
        sigma = flex.double(handle.sigma)
        rlp = flex.double(handle.rlp)
        peak = flex.double(handle.peak) * 0.01
        if len(handle.iseg):
            panel = flex.size_t(handle.iseg) - 1
        else:
            panel = flex.size_t(len(hkl), 0)
        Command.end(f"Read {len(hkl)} reflections from INTEGRATE.HKL file.")

        if len(self._experiment.detector) > 1:
            for p_id, p in enumerate(self._experiment.detector):
                sel = panel == p_id
                offset = p.get_raw_image_offset()
                xyzcal.set_selected(sel, xyzcal.select(sel) - (offset[0], offset[1], 0))
                xyzobs.set_selected(sel, xyzobs.select(sel) - (offset[0], offset[1], 0))

        # Derive the reindex matrix
        rdx = self.derive_reindex_matrix(handle)
        print("Reindex matrix:\n%d %d %d\n%d %d %d\n%d %d %d" % (rdx.elems))

        # Reindex the reflections
        Command.start("Reindexing reflections")
        cb_op = sgtbx.change_of_basis_op(sgtbx.rt_mx(sgtbx.rot_mx(rdx.elems)))
        hkl = cb_op.apply(hkl)
        Command.end(f"Reindexed {len(hkl)} reflections")

        # Create the reflection list
        Command.start("Creating reflection table")
        table = flex.reflection_table()
        table["id"] = flex.int(len(hkl), 0)
        table["panel"] = panel
        table["miller_index"] = hkl
        table["xyzcal.px"] = xyzcal
        table["xyzobs.px.value"] = xyzobs
        table["intensity.cor.value"] = iobs
        table["intensity.cor.variance"] = flex.pow2(sigma)
        table["intensity.prf.value"] = iobs * peak / rlp
        table["intensity.prf.variance"] = flex.pow2(sigma * peak / rlp)
        table["lp"] = 1.0 / rlp
        table["d"] = flex.double(uc.d(h) for h in hkl)
        Command.end(f"Created table with {len(table)} reflections")

        # Output the table to pickle file
        if params.output.filename is None:
            params.output.filename = "integrate_hkl.refl"
        Command.start(f"Saving reflection table to {params.output.filename}")
        table.as_file(params.output.filename)
        Command.end(f"Saved reflection table to {params.output.filename}")

    def derive_reindex_matrix(self, handle):
        """Derive a reindexing matrix to go from the orientation matrix used
        for XDS integration to the one used for DIALS integration."""
        dA = matrix.sqr(self._experiment.crystal.get_A())
        dbeam = matrix.col(self._experiment.beam.get_sample_to_source_direction())
        daxis = matrix.col(self._experiment.goniometer.get_rotation_axis())
        n = dbeam.cross(daxis)
        xbeam = matrix.col(handle.beam_vector).normalize()
        xaxis = matrix.col(handle.rotation_axis).normalize()

        # want to align XDS -s0 vector...
        R = align_reference_frame(-xbeam, dbeam, xaxis, n.cross(dbeam))
        xA = matrix.sqr(
            handle.unit_cell_a_axis + handle.unit_cell_b_axis + handle.unit_cell_c_axis
        ).inverse()
        xA = R * xA

        # assert that this should just be a simple integer rotation matrix
        # i.e. reassignment of a, b, c so...

        return matrix.sqr([int(round(e)) for e in (dA.inverse() * xA).elems])


class XDSFileImporter:
    """Import a data block from xds."""

    def __init__(self, args):
        """Initialise with the options"""
        self.args = args

    def __call__(self, params, options):
        # Get the XDS.INP file
        xds_inp = os.path.join(self.args[0], "XDS.INP")
        if params.input.xds_file is None:
            xds_file = XDSFileImporter.find_best_xds_file(self.args[0])
        else:
            xds_file = os.path.join(self.args[0], params.input.xds_file)

        # Check a file is given
        if xds_file is None:
            raise RuntimeError("No XDS file found")

        # Load the experiment list
        unhandled = []
        experiments = ExperimentListFactory.from_xds(xds_inp, xds_file)

        # Print out any unhandled files
        if len(unhandled) > 0:
            print("-" * 80)
            print("The following command line arguments were not handled:")
            for filename in unhandled:
                print(f"  {filename}")

        # Print some general info
        print("-" * 80)
        print(f"Read {len(experiments)} experiments from {xds_file}")

        # Attempt to create scan-varying crystal model if requested
        if params.read_varying_crystal:
            integrate_lp = os.path.join(self.args[0], "INTEGRATE.LP")
            if os.path.isfile(integrate_lp):
                self.extract_varying_crystal(integrate_lp, experiments)
            else:
                print("No INTEGRATE.LP to extract varying crystal model. Skipping")

        # Loop through the data blocks
        for i, exp in enumerate(experiments):

            # Print some experiment info
            print("-" * 80)
            print("Experiment %d" % i)
            print(f"  format: {exp.imageset.get_format_class()}")
            print(f"  type: {type(exp.imageset)}")
            print(f"  num images: {len(exp.imageset)}")

            # Print some model info
            if options.verbose > 1:
                print("")
                if exp.beam:
                    print(exp.beam)
                else:
                    print("no beam!")
                if exp.detector:
                    print(exp.detector)
                else:
                    print("no detector!")
                if exp.goniometer:
                    print(exp.goniometer)
                else:
                    print("no goniometer!")
                if exp.scan:
                    print(exp.scan)
                else:
                    print("no scan!")
                if exp.crystal:
                    print(exp.crystal)
                else:
                    print("no crystal!")

        # Write the experiment list to a JSON or pickle file
        if params.output.filename is None:
            params.output.filename = "xds_models.expt"
        print("-" * 80)
        print(f"Writing experiments to {params.output.filename}")
        experiments.as_file(params.output.filename)

        # Optionally save as a data block
        if params.output.xds_experiments:
            print("-" * 80)
            print(f"Writing data block to {params.output.xds_experiments}")
            experiments.as_file(params.output.xds_experiments)

    @staticmethod
    def find_best_xds_file(xds_dir):
        """Find the best available file."""

        # The possible files to check
        paths = [
            os.path.join(xds_dir, "XDS_ASCII.HKL"),
            os.path.join(xds_dir, "INTEGRATE.HKL"),
            os.path.join(xds_dir, "GXPARM.XDS"),
            os.path.join(xds_dir, "XPARM.XDS"),
        ]

        # Return the first path that exists
        for p in paths:
            if os.path.exists(p):
                return p

        # If no path exists, return None
        return None

    @staticmethod
    def extract_varying_crystal(integrate_lp, experiments):
        """Extract a varying crystal model from an INTEGRATE.LP file (static
        in blocks with step changes) and write it to the provided
        experiments
        """

        if len(experiments) > 1:
            print(
                "Can only read a varying crystal model for a single "
                + "experiment. Skipping."
            )
            return
        experiment = experiments[0]

        # read required records from the file. Relies on them being in the
        # right order as we read through once
        xds_axis = None
        xds_beam = None
        blocks, a_axis, b_axis, c_axis = [], [], [], []
        with open(integrate_lp) as f:
            for record in f:
                if record.lstrip().startswith("ROTATION_AXIS="):
                    xds_axis = record.split("ROTATION_AXIS=")[1].split()
                    break
            for record in f:
                if record.lstrip().startswith("INCIDENT_BEAM_DIRECTION="):
                    xds_beam = record.split("INCIDENT_BEAM_DIRECTION=")[1].split()
                    break
            for record in f:
                if record.lstrip().startswith("PROCESSING OF IMAGES"):
                    blocks.append(record.split("PROCESSING OF IMAGES")[1])
                    continue
                if record.lstrip().startswith("COORDINATES OF UNIT CELL A-AXIS"):
                    a_axis.append(record.split("COORDINATES OF UNIT CELL A-AXIS")[1])
                    continue
                if record.lstrip().startswith("COORDINATES OF UNIT CELL B-AXIS"):
                    b_axis.append(record.split("COORDINATES OF UNIT CELL B-AXIS")[1])
                    continue
                if record.lstrip().startswith("COORDINATES OF UNIT CELL C-AXIS"):
                    c_axis.append(record.split("COORDINATES OF UNIT CELL C-AXIS")[1])
                    continue

        # sanity checks
        msg = "INTEGRATE.LP is not in the expected format"
        nblocks = len(blocks)
        try:
            assert len(a_axis) == len(b_axis) == len(c_axis) == nblocks
            assert (xds_axis, xds_beam).count(None) == 0
        except AssertionError:
            print(msg)
            return

        # conversions to numeric
        try:
            blocks = [[int(b) for b in block.split("...")] for block in blocks]
            a_axis = [[float(a) for a in axis.split()] for axis in a_axis]
            b_axis = [[float(b) for b in axis.split()] for axis in b_axis]
            c_axis = [[float(c) for c in axis.split()] for axis in c_axis]
            xds_beam = [float(e) for e in xds_beam]
            xds_axis = [float(e) for e in xds_axis]
        except ValueError:
            print(msg)
            return

        # coordinate frame conversions
        dbeam = matrix.col(experiment.beam.get_sample_to_source_direction())
        daxis = matrix.col(experiment.goniometer.get_rotation_axis())
        xbeam = matrix.col(xds_beam).normalize()
        xaxis = matrix.col(xds_axis).normalize()

        # want to align XDS -s0 vector...
        R = align_reference_frame(-xbeam, dbeam, xaxis, daxis)

        # Make a static crystal for each block
        crystals = []
        sg = experiment.crystal.get_space_group()
        for a, b, c in zip(a_axis, b_axis, c_axis):
            a = R * matrix.col(a)
            b = R * matrix.col(b)
            c = R * matrix.col(c)
            crystals.append(Crystal(a, b, c, space_group=sg))

        # construct a list of scan points
        A_list = []
        for block, crystal in zip(blocks, crystals):
            A = crystal.get_A()
            for im in range(block[0], block[1] + 1):
                A_list.append(A)
        # Need a final scan point at the end of the final image
        A_list.append(A)

        # set the scan-varying crystal
        experiment.crystal.set_A_at_scan_points(A_list)


class Script:
    """A class to encapsulate the script."""

    def __init__(self):
        """Initialise the script."""
        # Create the phil parameters
        phil_scope = parse(
            """
      input {
        method = *experiment reflections
          .type = choice
          .help = "The input method"

        xds_file = None
          .type = str
          .help = "Explicitly specify the file to use"
      }

      output {
        filename = None
          .type = str
          .help = "The output file"

        xds_experiments = None
          .type = str
          .help = "Output filename of data block with xds"
      }

      remove_invalid = False
        .type = bool
        .help = "Remove non-index reflections (if miller indices are present)"

      add_standard_columns = False
        .type = bool
        .help = "Add empty standard columns to the reflections. Note columns"
                "for centroid variances are set to contain 1s, not 0s"

      read_varying_crystal = False
        .type = bool
        .help = "Attempt to create a scan-varying crystal model from"
                "INTEGRATE.LP, if present"
    """
        )

        # The option parser
        usage = "usage: dials.import_xds [options] (SPOT.XDS|INTEGRATE.HKL)"
        self.parser = ArgumentParser(usage=usage, phil=phil_scope)

    def run(self, args=None):
        """Run the script."""

        # Parse the command line arguments
        params, options, args = self.parser.parse_args(
            args, show_diff_phil=True, return_unhandled=True
        )

        # Check number of arguments
        if len(args) == 0:
            self.parser.print_help()
            exit(0)

        # Select the importer class
        if params.input.method == "experiment":
            importer = XDSFileImporter(args)
        else:
            importer = self.select_importer(args)

        # Import the XDS data
        importer(params, options)

    def select_importer(self, args):
        path, filename = os.path.split(args[0])
        if filename == "SPOT.XDS":
            return SpotXDSImporter(args[0])
        elif filename == "INTEGRATE.HKL":
            assert len(args) == 2
            experiments = ExperimentListFactory.from_json_file(args[1])
            assert len(experiments) == 1
            return IntegrateHKLImporter(args[0], experiments[0])
        else:
            raise RuntimeError(f"expected (SPOT.XDS|INTEGRATE.HKL), got {filename}")


@show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



"""Definition of systematic absences check algorithm."""
from __future__ import annotations

import logging

from cctbx import sgtbx

from dials.algorithms.symmetry.absences.laue_groups_info import (
    laue_groups,
    score_screw_axes,
    score_space_groups,
)
from dials.util import tabulate

logger = logging.getLogger("dials")


def run_systematic_absences_checks(
    experiments, merged_reflections, significance_level=0.95, method="direct"
):
    """Check for systematic absences in the data for the laue group.

    Using a reflection table containing merged data, test screw axes and score
    possible space groups. The crystals are updated with the most likely space
    group.
    """
    # Get the laue class from the space group.
    space_group = experiments[0].crystal.get_space_group()
    laue_group = str(space_group.build_derived_patterson_group().info())
    logger.info("Laue group: %s", laue_group)
    if laue_group not in laue_groups:
        logger.info("No absences to check for this laue group")
        return

    # Score the screw axes.
    screw_axes, screw_axis_scores = score_screw_axes(
        laue_groups[laue_group], merged_reflections, significance_level, method
    )

    logger.info(
        f"Scoring method: {method}\n"
        + tabulate(
            [
                [
                    a.name,
                    f"{score:.3f}",
                    str(a.n_refl_used[0]),
                    str(a.n_refl_used[1]),
                    f"{a.mean_I:.3f}",
                    f"{a.mean_I_abs:.3f}",
                    f"{a.mean_I_sigma:.3f}",
                    f"{a.mean_I_sigma_abs:.3f}",
                ]
                for a, score in zip(screw_axes, screw_axis_scores)
            ],
            [
                "Screw axis",
                "Score",
                "No. present",
                "No. absent",
                "<I> present",
                "<I> absent",
                "<I/sig> present",
                "<I/sig> absent",
            ],
        ),
    )

    # Score the space groups from the screw axis scores.
    space_groups, scores = score_space_groups(
        screw_axis_scores, laue_groups[laue_group]
    )

    logger.info(
        "%s",
        tabulate(
            [[sg, f"{score:.4f}"] for sg, score in zip(space_groups, scores)],
            ["Space group", "score"],
        ),
    )

    # Find the best space group and update the experiments.
    best_sg = space_groups[scores.index(max(scores))]
    logger.info("Recommended space group: %s", best_sg)
    if "enantiomorphic pairs" in laue_groups[laue_group]:
        if best_sg in laue_groups[laue_group]["enantiomorphic pairs"]:
            logger.info(
                "Space group with equivalent score (enantiomorphic pair): %s",
                laue_groups[laue_group]["enantiomorphic pairs"][best_sg],
            )

    new_sg = sgtbx.space_group_info(symbol=best_sg).group()
    for experiment in experiments:
        experiment.crystal.set_space_group(new_sg)



"""Trivial check for whether classification of reflections as exiting or
entering the Ewald sphere is done the right way round"""


from __future__ import annotations

import math

import pytest

from . import geometry_phil, minimiser_phil


def test():
    from cctbx.sgtbx import space_group, space_group_symbols

    # We will set up a mock scan
    from dxtbx.model import ScanFactory
    from dxtbx.model.experiment_list import Experiment, ExperimentList
    from libtbx.phil import parse
    from scitbx import matrix
    from scitbx.array_family import flex

    from dials.algorithms.refinement.prediction.managed_predictors import (
        ScansExperimentsPredictor,
        ScansRayPredictor,
    )

    # Reflection prediction
    from dials.algorithms.spot_prediction import IndexGenerator

    # Building experimental models
    from .setup_geometry import Extract

    master_phil = parse(f"{geometry_phil}\n{minimiser_phil}")

    overrides = """geometry.parameters.crystal.a.length.range = 10 50
  geometry.parameters.crystal.b.length.range = 10 50
  geometry.parameters.crystal.c.length.range = 10 50"""

    models = Extract(master_phil, local_overrides=overrides)

    mydetector = models.detector
    mygonio = models.goniometer
    mycrystal = models.crystal
    mybeam = models.beam

    #############################
    # Generate some reflections #
    #############################

    # All indices in a 2.0 Angstrom sphere
    resolution = 2.0
    index_generator = IndexGenerator(
        mycrystal.get_unit_cell(),
        space_group(space_group_symbols(1).hall()).type(),
        resolution,
    )
    indices = index_generator.to_array()

    # Build a mock scan for a 30 degree sequence
    sf = ScanFactory()
    myscan = sf.make_scan(
        image_range=(1, 300),
        exposure_times=0.1,
        oscillation=(0, 0.1),
        epochs=list(range(300)),
        deg=True,
    )
    sequence_range = myscan.get_oscillation_range(deg=False)
    assert sequence_range == pytest.approx((0.0, math.pi / 6.0))
    im_width = myscan.get_oscillation(deg=False)[1]
    assert im_width == pytest.approx(0.1 * math.pi / 180.0)

    # Create an ExperimentList for ScansRayPredictor
    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=mybeam,
            detector=mydetector,
            goniometer=mygonio,
            scan=myscan,
            crystal=mycrystal,
            imageset=None,
        )
    )

    # Select those that are excited in a 30 degree sequence and get angles
    ray_predictor = ScansRayPredictor(experiments, sequence_range)
    obs_refs = ray_predictor(indices)

    # Set the experiment number
    obs_refs["id"] = flex.int(len(obs_refs), 0)

    # Calculate intersections
    ref_predictor = ScansExperimentsPredictor(experiments)
    obs_refs = ref_predictor(obs_refs)

    print("Total number of observations made", len(obs_refs))

    s0 = matrix.col(mybeam.get_s0())
    spindle = matrix.col(mygonio.get_rotation_axis())

    for ref in obs_refs.rows():
        # get the s1 vector of this reflection
        s1 = matrix.col(ref["s1"])

        r = s1 - s0
        r_orig = r.rotate_around_origin(spindle, -1.0, deg=True)

        # is it outside the Ewald sphere (i.e. entering)?
        test = (s0 + r_orig).length() > s0.length()
        assert ref["entering"] == test
